import logging
from sys import exit

import pandas as pd

logging.getLogger(__name__)

__author__ = "Rayid Ghani, Pedro Saleiro <saleiro@uchicago.edu>"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."

def check_required_cols(df, required_cols):
    """

    :param df: A data frame of model results
    :param required_cols: Column names required for selected fairness measures
    :return: None, or ValueError
    """
    check_model_cols = [col in df.columns for col in required_cols]
    if False in check_model_cols:
        raise ValueError
    return


def get_attr_cols(df, non_attr_cols):
    """
    :param df: A data frame of model results
    :param non_attr_cols: Names of columns not associated with attributes
    :return: List of columns associated with sample attributes
    """
    # index of the columns that are associated with attributes
    attr_cols = df.columns[~df.columns.isin(non_attr_cols)]
    if attr_cols.empty:
        raise ValueError
    return attr_cols.tolist()


def discretize(df, target_cols):
    """

    :param df: A data frame of model results
    :param target_cols: Names of columns to discretize
    :return: A data frame
    """
    for col in target_cols:
        if len(df[col].unique()) > 1:
            bins, values = pd.qcut(df[col], 4, precision=2, labels=False, duplicates='drop', retbins=True)
            try:
                df[col] = bins.map(lambda x: '%0.2f' % values[x] + '-' + '%0.2f' % values[x + 1])
            except Exception as e:
                logging.info('Something strange with a column in the input_df ' + str(e))
                df = df.drop(col)
        else:
            try:
                df[col] = df[col].astype(str)
            except Exception as e:
                logging.info('Something strange with a column in the input_df ' + str(e))
                df = df.drop(col)
    return df


def preprocess_input_df(df, required_cols=None):
    """

    :param df: A data frame of model results
    :param required_cols: Names of columns required for bias calculations.
        Default is None.
    :return: A data frame, list of columns associated with sample attributes
    """
    if not required_cols:
        required_cols = ['score']
    try:
        check_required_cols(df, required_cols)
    except ValueError:
        logging.error('preprocessing.preprocess_input_df: input dataframe does not have all the required columns.')
        exit(1)
    non_attr_cols = required_cols + ['model_id', 'as_of_date', 'entity_id', 'rank_abs', 'rank_pct', 'id', 'label_value']
    non_string_cols = df.columns[(df.dtypes != object) & (df.dtypes != str) & (~df.columns.isin(non_attr_cols))]
    df = discretize(df, non_string_cols)
    try:
        attr_cols_input = get_attr_cols(df, non_attr_cols)
    except ValueError:
        logging.info('preprocessing.preprocess_input_df: input dataframe does not have any other columns besides required '
                      'columns. Please add attribute columns to the input df.')
        exit(1)
    return df, attr_cols_input



import logging
import os

import yaml

logging.getLogger(__name__)

__author__ = "Rayid Ghani, Pedro Saleiro <saleiro@uchicago.edu>, Loren Hinkson"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."

class Configs(object):
    original_fairness_measures = (
        'Statistical Parity', 'Impact Parity', 'FDR Parity',
        'FPR Parity', 'FNR Parity', 'FOR Parity', 'TPR Parity',
        'Precision Parity')

    def __init__(self, ref_groups_method='majority', fairness_threshold=0.8,
                 attr_cols=None, report=True, score_thresholds=None,
                 project_description={'title': '', 'goal': ''},
                 ref_groups=None, db=None,
                 fairness_measures=original_fairness_measures,
                 plot_bias_disparities=(), plot_bias_metrics=()):

        self.ref_groups_method = ref_groups_method
        self.fairness_threshold = fairness_threshold
        self.attr_cols = attr_cols
        self.report = report
        self.score_thresholds = score_thresholds
        self.ref_groups = ref_groups
        self.db = db
        self.fair_measures_requested = list(fairness_measures)
        self.project_description = project_description
        self.plot_bias_metrics = plot_bias_metrics
        self.plot_bias_disparities = plot_bias_disparities

    @staticmethod
    def load_configs(configs_path):
        try:
            if configs_path:
                with open(configs_path, 'r') as stream:
                    configs_fromfile = yaml.load(stream)
            else:
                configs_fromfile = {}
        except FileNotFoundError:
            logging.error('**Could not find configurations! Please set path to configs.yaml file using --config')
            print(configs_path)
            print(os.getcwd())
            exit(1)
        return Configs(**configs_fromfile)



import datetime
import logging

import pandas as pd
from markdown2 import markdown
from tabulate import tabulate

tabulate.PRESERVE_WHITESPACE = True

logging.getLogger(__name__)

__author__ = "Rayid Ghani, Pedro Saleiro <saleiro@uchicago.edu>, Loren Hinkson"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."

#
#  DISCLAIMER: these methods were developed with a particular version of the webapp in mind. They lack flexibility (lots of
# hardcoded things)!!!
#             If new features(fairness measures, etc..) are added to the webapp this needs to change a lot...
#

def get_group_value_report(group_value_df):
    """

    :param group_value_df:
    :return:
    """
    group_value_report = {}
    the_false_df = group_value_df.loc[(group_value_df['Unsupervised Fairness'] == False) | (
            group_value_df['Supervised Fairness'] == False)]
    for index, row in the_false_df.iterrows():
        report = ''
        metrics = []
        group = row['group_variable'] + ' = ' + row['group_value']
        text1 = group + ' does not have '
        report += text1
        text2 = ''
        if row['Unsupervised Fairness'] is False:
            text2 += 'Unsupervised Fairness '
            text3 = ''
            if row['Statistical Parity'] is False:
                text3 += '(no Statistical Parity'
                ref_val = 0.0
                ref_group_value = group_value_df.loc[(group_value_df['group_variable'] == row[
                    'group_variable']) & (group_value_df['group_value'] == row[
                    'ppr_ref_group_value'])]['ppr'].values[0]
                ppr_text = '{:.0f}% of the group is selected, compared to {:.0f} % of the ' \
                           'reference group '.format(row['ppr'] * 100, ref_group_value * 100) + \
                           row['group_variable'] + ' = ' + row['ppr_ref_group_value']
                metrics.append(ppr_text)
            if row['Impact Parity'] is False:
                if text3 == '':
                    text3 += '(no Impact Parity)'
                else:
                    text3 += ', no Impact Parity)'
                pprev_text = ''
            else:
                text3 += ')'
            text2 += text3
        if row['Supervised Fairness'] is False:
            if text2 != '':
                text2 += ' neither '
            text2 += 'Supervised Fairness '
            text3 = ''
            if row['TypeI Parity'] is False:
                text3 += '(no Type I Parity'
            if row['TypeII Parity'] is False:
                if text3 == '':
                    text3 += '(no Type II Parity)'
                else:
                    text3 += ', no Type II Parity)'
            else:
                text3 += ') '
            text2 += text3
        report += text2
        group_value_report[group] = [report, metrics]

    return group_value_report


def get_highlevel_report(group_attribute_df):
    group_attribute_df = group_attribute_df.applymap(str)
    cols = ['attribute_name']
    if 'Unsupervised Fairness' in group_attribute_df.columns:
        cols.append('Unsupervised Fairness')
    if 'Supervised Fairness' in group_attribute_df.columns:
        cols.append('Supervised Fairness')
    group_attribute_df = group_attribute_df[cols]
    map = {}
    attr_list = group_attribute_df['attribute_name'].unique()
    for col in group_attribute_df.columns:
        if col == 'attribute_name':
            colstr = 'Attribute'
        else:
            colstr = col
        map[col] = colstr
        # to be able to click on true/false and redirect to the next section
        if col != 'attribute_name':
            for attr in attr_list:
                group_attribute_df.loc[group_attribute_df['attribute_name'] == attr, col] = '[' + group_attribute_df[col][
                    group_attribute_df['attribute_name'] == attr] + ']' + '(#' + '-'.join(attr.lower().split(' ')) + ')'
    for attr in attr_list:
        group_attribute_df = group_attribute_df.replace(attr, '[' + attr + ']' + '(#' + '-'.join(attr.lower().split(' ')) + ')')
    group_attribute_df = group_attribute_df.rename(index=str, columns=map)
    highlevel_report = tabulate(group_attribute_df, headers='keys', tablefmt='pipe', showindex='never', numalign="left")
    return highlevel_report


def get_parity_group_report(group_value_df, attribute, fairness_measures, fairness_measures_depend):
    group_value_df = group_value_df.round(2)
    group_value_df = group_value_df.applymap(str)
    def_cols = ['attribute_value']
    aux_df = group_value_df.loc[group_value_df['attribute_name'] == attribute]
    metrics = {}
    for par, disp in fairness_measures_depend.items():
        if par in fairness_measures:
            metrics[par] = disp

    # getting a reference group label
    for col in aux_df.columns:
        if col in metrics.keys():
            ref_group = metrics[col].replace('_disparity', '_ref_group_value')
            idx = aux_df.loc[aux_df['attribute_value'] == aux_df[ref_group]].index
            aux_df.loc[idx, col] = 'Ref'

    map = {}
    aux_df = aux_df[def_cols + fairness_measures]
    for col in aux_df.columns:
        if col == 'attribute_value':
            colstr = 'Attribute Value'
        else:
            colstr = col
        map[col] = colstr  #+ ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
        aux_df[col] = '[' + aux_df[col] + ']' + '(#' + '-'.join(attribute.lower().split(' ')) + '-2)'
    aux_df = aux_df.rename(index=str, columns=map)
    cols_order = ['Attribute Value', 'Statistical Parity', 'Impact Parity', 'FDR Parity', 'FPR Parity', 'FOR Parity',
                  'FNR Parity']
    new_order = [col for col in cols_order if col in aux_df.columns]
    aux_df = aux_df[new_order]

    parity_group = tabulate(aux_df,
                            headers='keys',
                            tablefmt='pipe', showindex='never', numalign="left")
    return parity_group


def setup_group_value_df(group_value_df, fairness_measures, fairness_measures_depend):
    group_value_df = group_value_df.round(2)
    group_value_df = group_value_df.applymap(str)
    group_size = group_value_df['group_size_pct']
    metrics = {}
    for par, disp in fairness_measures_depend.items():
        if par in fairness_measures:
            metrics[disp] = par
            metrics[disp.replace('_disparity', '')] = par
    aux_df = group_value_df[['attribute_name', 'attribute_value'] + list(metrics.keys())]
    for col in group_value_df.columns:
        if col in metrics.keys():
            # we want to keep the ref group without green/red so we need to know the name of the column to search for
            if not col.endswith('_disparity'):
                ref_group = col + '_ref_group_value'
            else:
                ref_group = col.replace('_disparity', '_ref_group_value')
            group_value_df.loc[(group_value_df[metrics[col]] == 'True') & (group_value_df['attribute_value'] != group_value_df[
                ref_group]), col] = '##green## ' + group_value_df[col][group_value_df[metrics[col]] == 'True']

            group_value_df.loc[group_value_df[metrics[col]] == 'False', col] = '##red##' + group_value_df[col][group_value_df[
                                                                                                                   metrics[
                                                                                                                       col]] == 'False']
    group_value_df['group_size_pct'] = group_size
    print('******************', group_value_df['group_size_pct'])
    return group_value_df


def get_disparities_group_report(group_value_df, attribute, fairness_measures, fairness_measures_depend):
    def_cols = ['attribute_value']
    metrics = {}
    for par, disp in fairness_measures_depend.items():
        if par in fairness_measures:
            metrics[disp] = par
    aux_df = group_value_df.loc[group_value_df['attribute_name'] == attribute]
    aux_df = aux_df[def_cols + list(metrics.keys())]
    map = {}
    for col in aux_df.columns:
        colstr = col.replace('_', ' ')
        if col == 'attribute_value':
            colstr = 'Attribute Value'
        else:
            colstr = colstr.split(' ')[0].upper() + ' Disparity'
        map[col] = colstr  #+ ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
        aux_df[col] = '[' + aux_df[col] + ']' + '(#' + '-'.join(attribute.lower().split(' ')) + '-3)'
    aux_df = aux_df.rename(index=str, columns=map)
    # this is hardcoded. If metrics supported by aequitas change this needs to change
    cols_order = ['Attribute Value', 'PPR Disparity', 'PPREV Disparity', 'FDR Disparity', 'FPR Disparity', 'FOR Disparity',
                  'FNR Disparity']
    new_order = [col for col in cols_order if col in aux_df.columns]
    aux_df = aux_df[new_order]
    disparities_group = tabulate(aux_df,
                                 headers='keys',
                                 tablefmt='pipe', showindex='never', numalign="left")

    return disparities_group


def get_group_group_report(group_value_df, attribute, fairness_measures, fairness_measures_depend):
    # defining how to display stuff
    names = {'attribute_value': 'Attribute Value',
             'group_size_pct': 'Group Size Ratio'}
    def_cols = ['attribute_value', 'group_size_pct']
    for par, disp in fairness_measures_depend.items():
        if par in fairness_measures:
            def_cols.append(disp.replace('_disparity', ''))
    aux_df = group_value_df.loc[group_value_df['attribute_name'] == attribute]
    aux_df = aux_df[def_cols]
    aux_df = aux_df.round(2)
    aux_df = aux_df.astype(str)
    # fixing the same order of columns every time!
    cols_order = ['attribute_value', 'group_size_pct', 'ppr', 'pprev', 'fdr', 'fpr', 'for', 'fnr']
    new_order = [col for col in cols_order if col in aux_df.columns]
    aux_df = aux_df[new_order]
    map = {}
    for col in aux_df.columns:
        if col in names:
            colstr = names[col]
        else:
            colstr = col.upper()
        map[col] = colstr  #+ ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
    aux_df = aux_df.rename(index=str, columns=map)
    group_group = tabulate(aux_df,
                           headers='keys',
                           tablefmt='pipe', showindex='never', numalign="left")
    return group_group


def get_sentence_highlevel(fair_results):
    sent = '**The Bias Report audited the risk assessmentt system and has found that is has'
    is_fair = ' passed' if fair_results['Overall Fairness'] is True else ' failed'
    sent += is_fair + ' the audit with respect to the following fairness criteria:**\n\n'
    return sent



def get_statpar_text(group_value_df, fairness_measures_depend):
    group_value_df = group_value_df.round(2)
    group_value_df = group_value_df.applymap(str)
    fairness_metric = 'Statistical Parity'
    false_df = group_value_df.loc[group_value_df[fairness_metric] == 'False']
    bias_metric = fairness_measures_depend[fairness_metric]
    group_metric = bias_metric.replace('_disparity', '')
    ref_group_col = group_metric + '_ref_group_value'
    text_detail = ''
    false_dict = {}
    for index, row in false_df.iterrows():
        ref_group_row = group_value_df.loc[(group_value_df['attribute_name'] == row['attribute_name']) &
                                           (group_value_df['attribute_value'] == row[ref_group_col])]
        sentence = ' is' \
                   ' **{group_metric_value}**% of positive class.' \
                   '' \
                   ''.format(
            group_metric_value='%.0f' % (float(row[group_metric]) * 100),
        )

        try:
            false_dict[row['attribute_name']].add('[' + row['attribute_value'] + '](#equal-parity)' + sentence)
        except KeyError:
            false_dict[row['attribute_name']] = set()
            false_dict[row['attribute_name']].add('[' + row['attribute_value'] + '](#equal-parity)' + sentence)

    if false_df.empty:
        cellref = '##green##Based on the fairness threshold used, the number of selected positives is similar across ' \
                  'different ' \
                  'groups.\n\n'
    else:
        cellref = ''
        for key in false_dict.keys():
            cellref += '**{attribute_name}:** ##br##&emsp;&emsp;&emsp;'.format(attribute_name=key)
            cellref += '##br##&emsp;&emsp;&emsp;'.join(false_dict[key]) + ' ##br##'

    return cellref


def get_impact_text(group_value_df, fairness_measures_depend):
    group_value_df = group_value_df.round(2)
    group_value_df = group_value_df.applymap(str)
    fairness_metric = 'Impact Parity'
    false_df = group_value_df.loc[group_value_df[fairness_metric] == 'False']
    bias_metric = fairness_measures_depend[fairness_metric]
    group_metric = bias_metric.replace('_disparity', '')
    ref_group_col = group_metric + '_ref_group_value'
    text_detail = ''
    false_dict = {}
    for index, row in false_df.iterrows():
        ref_group_row = group_value_df.loc[(group_value_df['attribute_name'] == row['attribute_name']) &
                                           (group_value_df['attribute_value'] == row[ref_group_col])]

        sentence = ': **{group_metric_value}**% of the group is in the selected set (classified as positive),' \
                   ' in comparison to {ref_group_metric_value}% from the reference group \"{' \
                   'ref_group_value}\"' \
                   ''.format(
            group_metric_value='%.0f' % (float(row[group_metric]) * 100),
            attribute_value=row['attribute_value'],
            ref_group_metric_value='%.0f' % (float(ref_group_row[group_metric].values[0]) * 100),
            ref_group_value=row[ref_group_col])

        try:
            false_dict[row['attribute_name']].add('[' + row['attribute_value'] + '](#proportional-parity)' + sentence)
        except KeyError:
            false_dict[row['attribute_name']] = set()
            false_dict[row['attribute_name']].add('[' + row['attribute_value'] + '](#proportional-parity)' + sentence)

    if false_df.empty:
        cellref = '##green##Based on the fairness threshold used, the percentage of selected individuals from ' \
                  'each group is not disparate to the percentage of selected individuals of the respective reference group.\n\n'
    else:
        cellref = ''
        for key in false_dict.keys():
            cellref += '**{attribute_name}:** ##br##&emsp;&emsp;&emsp;'.format(attribute_name=key)
            cellref += '##br##&emsp;&emsp;&emsp;'.join(false_dict[key]) + ' ##br##'

    return cellref


def get_old_false_text(group_value_df, fairness_metric, fairness_measures_depend):
    names = {

        'fpr': 'false positive rate',
        'fnr': 'false negative rate',
        'fdr': 'false discovery rate',
        'for': 'false omission rate',
        'ppr': 'predicted positive ratio',
        'pprev': 'predicted prevalence (percentage of positives in the group)'
    }
    group_value_df = group_value_df.round(2)
    group_value_df = group_value_df.applymap(str)
    false_df = group_value_df.loc[group_value_df[fairness_metric] == 'False']
    bias_metric = fairness_measures_depend[fairness_metric]
    group_metric = bias_metric.replace('_disparity', '')
    ref_group_col = group_metric + '_ref_group_value'
    text_detail = ''
    false_dict = {}
    for index, row in false_df.iterrows():
        ref_group_row = group_value_df.loc[(group_value_df['attribute_name'] == row['attribute_name']) &
                                           (group_value_df['attribute_value'] == row[ref_group_col])]
        sentence = ': **{bias_metric_value}**% ' \
                   'of the {group_metric_name} of the reference group \"{ref_group_value}\",' \
                   ' corresponding to a difference of {group_metric_value} vs {ref_group_metric_value}.' \
            .format(
            group_metric_name=names[group_metric],
            bias_metric_value='%.0f' % (float(row[bias_metric]) * 100),
            ref_group_value=row[ref_group_col],
            group_metric_value=row[group_metric],
            ref_group_metric_value=ref_group_row[group_metric].values[0])
        try:
            false_dict[row['attribute_name']].add('[' + row['attribute_value'] + '](#false-positive-parity)' + sentence)
        except KeyError:
            false_dict[row['attribute_name']] = set()
            false_dict[row['attribute_name']].add('[' + row['attribute_value'] + '](#false-positive-parity)' + sentence)

    if false_df.empty:
        cellref = '##green##Based on the fairness threshold used, the percentage of selected elements from ' \
                  'each group is not disparate to the percentage of selected elements of the respective reference group.\n\n'
    else:
        cellref = ''
        for key in false_dict.keys():
            cellref += '**{attribute_name}:** ##br##&emsp;&emsp;&emsp;'.format(attribute_name=key)
            cellref += '##br##&emsp;&emsp;&emsp;'.join(false_dict[key]) + ' ##br##'

    return cellref


def get_false_text(group_value_df, fairness_metric, fairness_measures_depend):
    names = {

        'fpr': 'false positive rate',
        'fnr': 'false negative rate',
        'fdr': 'false discovery rate',
        'for': 'false omission rate',
        'ppr': 'predicted positive ratio',
        'pprev': 'predicted prevalence (percentage of positives in the group)'
    }
    group_value_df = group_value_df.round(2)
    group_value_df = group_value_df.applymap(str)
    false_df = group_value_df.loc[group_value_df[fairness_metric] == 'False']
    bias_metric = fairness_measures_depend[fairness_metric]
    group_metric = bias_metric.replace('_disparity', '')
    ref_group_col = group_metric + '_ref_group_value'
    text_detail = ''
    false_dict = {}
    ref_group_dict = {}
    for index, row in false_df.iterrows():
        ref_group_row = group_value_df.loc[(group_value_df['attribute_name'] == row['attribute_name']) &
                                           (group_value_df['attribute_value'] == row[ref_group_col])]
        sentence = ' with ++span-red-init++{bias_metric_value}X++span-red-end++ Disparity'.format(
            bias_metric_value='%.2f' % float(row[bias_metric]))

        ref_group_dict[row['attribute_name']] = ' (with reference group as **' + row[ref_group_col] + '**)'


        sentence2 = '{group_metric_name} of this group is {group_metric_value} compared to {ref_group_metric_value} for the reference group {ref_group_value}.' \
            .format(
            group_metric_name=names[group_metric],
            ref_group_value=row[ref_group_col],
            group_metric_value=row[group_metric],
            ref_group_metric_value=ref_group_row[group_metric].values[0])

        try:
            false_dict[row['attribute_name']].add('##tooltip-start-title##' + sentence2 + '##tooltip-end-title##' + row[
                'attribute_value'] + '##tooltip-end-anchor##' +
                                                  sentence)
        except KeyError:
            false_dict[row['attribute_name']] = set()
            false_dict[row['attribute_name']].add('##tooltip-start-title##' + sentence2 + '##tooltip-end-title##' + row[
                'attribute_value'] + '##tooltip-end-anchor##' +
                                                  sentence)
    if false_df.empty:
        cellref = '++span-green-init++Based on the fairness threshold used, all groups passed the audit for this metric.++span-green-end++\n\n'
    else:
        cellref = ''
        for key in false_dict.keys():
            cellref += '**For {attribute_name}**'.format(attribute_name=key) + ref_group_dict[key] + '##br##&emsp;&emsp;&emsp;'
            cellref += '##br##&emsp;&emsp;&emsp;'.join(false_dict[key]) + ' ##br##  ##br##'

    return cellref


def get_highlevel_table(group_value_df, fairness_measures, ):
    supported_name = {'Statistical Parity': '[Equal Parity](#equal-parity)',
                      'Impact Parity': '[Proportional Parity](#proportional-parity)',
                      'TypeI Parity': '[False Positive Parity](#false-positive-parity)',
                      'TypeII Parity': '[False Negative Parity](#false-negative-parity)'}
    supported_outcome = {'Statistical Parity': 'Each group is represented equally.',
                         'Impact Parity': 'Each group is represented proportional to their representation in the overall population.',
                         'TypeI Parity': 'Each group has proportionally equal false positive errors made by the model.',
                         'TypeII Parity': 'Each group has proportionally equal false negative errors made by the model.'}
    supported_order = ['Statistical Parity', 'Impact Parity', 'TypeI Parity', 'TypeII Parity']
    # once again this is hardcoded because it's easy now, in the future make it mapped automatically

    map_ref_groups = {'Statistical Parity': ['ppr_ref_group_value'],
                      'Impact Parity': ['pprev_ref_group_value'],
                      'TypeI Parity': ['fpr_ref_group_value', 'fdr_ref_group_value'],
                      'TypeII Parity': ['fnr_ref_group_value', 'for_ref_group_value']}
    key_columns = ['model_id', 'score_threshold', 'attribute_name']
    fairness_measures_edited = []
    for meas in fairness_measures:
        if meas in ['FPR Parity', 'FDR Parity']:
            fairness_measures_edited.append('TypeI Parity')
        elif meas in ['FNR Parity', 'FOR Parity']:
            fairness_measures_edited.append('TypeII Parity')
        else:
            fairness_measures_edited.append(meas)
    fairness_measures_edited = set(fairness_measures_edited)
    raw = {
        'Fairness Criteria': [],
        'Desired Outcome': [],
        'Reference Groups Selected': [],
        'Unfairly Affected Groups': []
    }
    for measure in supported_order:
        if measure in fairness_measures_edited:
            raw['Fairness Criteria'].append(supported_name[measure])
            raw['Desired Outcome'].append(supported_outcome[measure])
            false_df = group_value_df.loc[group_value_df[measure] == False]
            ref_dict = {}
            false_dict = {}
            for index, row in false_df.iterrows():
                try:
                    false_dict[row['attribute_name']].add('[' + row['attribute_value'] + ']' + \
                                                          supported_name[measure][supported_name[measure].find('('):])
                except KeyError:
                    false_dict[row['attribute_name']] = set()
                    false_dict[row['attribute_name']].add('[' + row['attribute_value'] + ']' + \
                                                          supported_name[measure][supported_name[measure].find('('):])
            attr_order = []
            if len(false_dict) > 0:
                cell = ''
                attr_order = false_dict.keys()
                for key in attr_order:
                    cell += '**{attribute_name}:**'.format(attribute_name=key)
                    cell += '##br##&emsp;&emsp;&emsp;'
                    cell += '##br##&emsp;&emsp;&emsp;'.join(false_dict[key]) + ' ##br##'
                raw['Unfairly Affected Groups'].append(cell)
            else:
                if group_value_df[measure].isnull().all():
                    raw['Unfairly Affected Groups'].append('Undefined')
                else:
                    raw['Unfairly Affected Groups'].append('No Unfair Groups Found')

            for ref in map_ref_groups[measure]:
                groupby_refs = group_value_df.groupby(key_columns + [ref])
                for group, values in groupby_refs:
                    try:
                        ref_dict[group[key_columns.index('attribute_name')]].add('[' + group[-1] + '](' + '-'.join(
                            supported_name[
                                measure]
                                .lower().split(' ')) + ')')
                    except KeyError:
                        ref_dict[group[key_columns.index('attribute_name')]] = set()
                        ref_dict[group[key_columns.index('attribute_name')]].add('[' + group[-1] + '](' + '-'.join(
                            supported_name[
                                measure].lower().split(
                                ' ')) + ')')
            cellref = ''
            align_rows = True if attr_order else False
            refs_order = attr_order if attr_order else ref_dict.keys()
            newline = '##br##'
            idented = '&emsp;&emsp;&emsp;'
            for key in refs_order:
                cellref += '**{attribute_name}:**'.format(attribute_name=key) + newline
                cellref += idented + list(ref_dict[key])[0] + ' ##br##'
                if align_rows:
                    cellref += ''.join([newline] * (len(false_dict[key]) - 1))  # this is the number of lines to skip in cell
                else:
                    cellref += newline
            raw['Reference Groups Selected'].append(cellref)

    highlevel_table = '\n\n'
    if len(raw['Fairness Criteria']) > 0:
        landf = pd.DataFrame(raw, columns=['Fairness Criteria', 'Desired Outcome', 'Reference Groups Selected',
                                           'Unfairly Affected Groups'])
        # keep the same order!!
        # krrp
        highlevel_table = tabulate(landf[['Fairness Criteria', 'Desired Outcome', 'Reference Groups Selected',
                                          'Unfairly Affected Groups']], headers='keys',
                                   tablefmt='pipe', showindex='never', numalign="left")

    return highlevel_table



def audit_summary(configs, group_value_df):
    supported_name2 = {'Statistical Parity': 'Equal Parity',
                       'Impact Parity': 'Proportional Parity',
                       'FPR Parity': 'False Positive Rate Parity',
                       'FDR Parity': 'False Discovery Rate Parity',
                       'FNR Parity': 'False Negative Rate Parity',
                       'FOR Parity': 'False Omission Rate Parity'}

    supported_name = {'Statistical Parity': '**Equal Parity** - Ensure all protected groups are have equal representation in the selected set.',
                      'Impact Parity': '**Proportional Parity** - Ensure all protected groups are selected proportional to their '
                                       'percentage of the population.',
                      'FPR Parity': '**False Positive Rate Parity** - Ensure all protected groups have the same false positive '
                                    'rates as the reference group).',
                      'FDR Parity': '**False Discovery Rate Parity** - Ensure all protected groups have equally proportional false '
                                    'positives within the selected set (compared to the reference group).',
                      'FNR Parity': '**False Negative Rate Parity** - Ensure all protected groups have the same false negative rates ('
                                    'as the reference group).',
                      'FOR Parity': '**False Omission Rate Parity** - Ensure all protected groups have equally proportional false '
                                    'negatives within the non-selected set (compared to the reference group).'}

    raw = {
        'column1': [],
        'column2': [],
        'column3': []
    }
    supported_order = ['Statistical Parity', 'Impact Parity', 'FPR Parity', 'FDR Parity', 'FNR Parity', 'FOR Parity']
    measures_results_dict = {}
    for measure in supported_order:
        if measure in configs.fair_measures_requested:
            raw['column1'].append(supported_name[measure])
            false_df = group_value_df.loc[group_value_df[measure] == False]
            if false_df.empty:
                link = ' [Details](#' + '-'.join(supported_name2[measure].lower().split(' ')) + \
                       '-span-green-initpassedspan-green-end)'
                raw['column2'].append('++span-green-init++**Passed**++span-green-end++')
                raw['column3'].append(link)

                measures_results_dict[measure] = '++span-green-init++Passed++span-green-end++'
            else:
                link = ' [Details](#' + '-'.join(supported_name2[measure].lower().split(' ')) + \
                       '-span-red-initfailedspan-red-end)'
                raw['column2'].append('++span-red-init++**Failed**++span-red-end++')
                raw['column3'].append(link)
                measures_results_dict[measure] = '++span-red-init++Failed++span-red-end++'
    df = pd.DataFrame(raw, columns=['column1', 'column2', 'column3'])
    summ_table = tabulate(df[['column1', 'column2', 'column3']], headers='keys',
                          tablefmt='pipe', showindex='never', numalign="left")
    return summ_table, measures_results_dict


def audit_description(configs, group_value_df):

    supported_name = {'Statistical Parity': '**Equal Parity** - Ensure all protected groups are have equal representation in the selected set.',
                      'Impact Parity': '**Proportional Parity** - Ensure all protected groups are selected proportional to their '
                                       'percentage of the population.',
                      'FPR Parity': '**False Positive Rate Parity** - Ensure all protected groups have the same false positive '
                                    'rates as the reference group).',
                      'FDR Parity': '**False Discovery Rate Parity** - Ensure all protected groups have equally proportional false '
                                    'positives within the selected set (compared to the reference group).',
                      'FNR Parity': '**False Negative Rate Parity** - Ensure all protected groups have the same false negative rates ('
                                    'as the reference group).',
                      'FOR Parity': '**False Omission Rate Parity** - Ensure all protected groups have equally proportional false '
                                    'negatives within the non-selected set (compared to the reference group).'}

    supported_order = ['Statistical Parity', 'Impact Parity', 'FPR Parity', 'FDR Parity', 'FNR Parity', 'FOR Parity']

    ref_group = {'predefined': 'Custom group - The reference groups you selected for each attribute will be used  to '
                               'calculate relative disparities in this audit.',
                 'majority': 'Majority group - The largest groups on each attribute will be used as baseline to calculate '
                             'relative '
                             'disparities in this audit.',
                 'min_metric': 'Automatically select, for each bias metric, the group on each attribute that has the '
                               'lower '
                               'value, to be used as baseline to calculate relative disparities in this audit.'
                 }

    raw = {
        'column1': [],
        'column2': []
    }

    raw['column1'].append('**Audit Date:**')
    raw['column2'].append(datetime.date.today().strftime('%d %b %Y'))
    raw['column1'].append('**Data Audited:**')
    raw['column2'].append('{:.0f} rows'.format(group_value_df['total_entities'].values[0]))
    raw['column1'].append('**Attributes Audited:**')
    raw['column2'].append(', '.join(group_value_df['attribute_name'].unique()))
    raw['column1'].append('**Audit Goal(s):**')
    measures = [supported_name[m] for m in supported_order if m in configs.fair_measures_requested]
    raw['column2'].append('\n'.join(measures) + '\n')
    raw['column1'].append('**Reference Groups:**')

    raw['column2'].append(ref_group[configs.ref_groups_method])
    raw['column1'].append('**Fairness Threshold:**')
    thresh = '{:.0f}%. If disparity for a group is within {' \
             ':.0f}% and ' \
             '{' \
             ':.0f}% of ' \
             ' the value of the reference group on a group metric (e.g. False ' \
             'Positive Rate), this audit will pass. ' \
             ''.format(
        float(configs.fairness_threshold) * 100, float(configs.fairness_threshold) * 100,
        float(1.0 / configs.fairness_threshold) * 100)
    raw['column2'].append(thresh)

    df = pd.DataFrame(raw, columns=['column1', 'column2'])
    desc_table = tabulate(df[['column1', 'column2']], headers='keys',
                          tablefmt='pipe', showindex='never', numalign="left")

    return desc_table

def audit_report_markdown(configs, group_value_df, fairness_measures_depend, overall_fairness, model_id=1):
    manylines = '\n\n&nbsp;\n\n&nbsp;\n\n'
    oneline = ' \n\n&nbsp;\n\n'
    mkdown_highlevel = '# The Bias Report'
    mkdown_highlevel += oneline
    mkdown_highlevel += audit_description(configs, group_value_df) + oneline + '----' + oneline

    # mkdown_highlevel += get_sentence_highlevel(overall_fairness) + oneline
    # mkdown_highlevel += get_highlevel_table(group_value_df, configs.fair_measures_requested) + oneline + '----' + oneline

    mkdown_highlevel += '### Audit Results:\n\n'
    mkdown_highlevel += '1. [Summary](#audit-results-summary)\n\n'
    mkdown_highlevel += '2. [Details by Fairness Measures](#audit-results-details-by-fairness-measures)\n\n'
    mkdown_highlevel += '3. [Details by Protected Attributes](#audit-results-details-by-protected-attributes)\n\n'
    mkdown_highlevel += '4. [Bias Metrics Values](#audit-results-bias-metrics-values)\n\n'
    mkdown_highlevel += '5. [Base Metrics Calculated for Each Group](#audit-results-group-metrics-values)\n\n' + oneline + '----' + \
                        oneline

    mkdown_highlevel += '### Audit Results: Summary' + '\n\n'
    summ_table, measures_results_dict = audit_summary(configs, group_value_df)
    mkdown_highlevel += summ_table + oneline + '----' + oneline

    mkdown_highlevel += '### Audit Results: Details by Fairness Measures' + oneline

    if 'Statistical Parity' in group_value_df.columns:
        mkdown_highlevel += '\n\n#### Equal Parity: ' + measures_results_dict['Statistical Parity'] + '\n\n'
        raw = {}
        raw['What is it?'] = ['##border##This criteria considers an attribute to have equal parity is every group is equally '
                              'represented in the selected set. For example, if race (with possible values of white, black, other) '
                              'has equal parity, it implies that all three races are equally represented (33% each)'
                              'in the selected/intervention set.']
        raw['When does it matter?'] = ['##border##If your desired ' \
                                       'outcome is to intervene equally on people ' \
                                                         'from all races, then you care about this criteria.']
        raw['Which groups failed the audit:'] = '##border##' + get_false_text(group_value_df, 'Statistical Parity',
                                                                              fairness_measures_depend)
        dft = pd.DataFrame(raw)
        mkdown_highlevel += tabulate(dft[['What is it?', 'When does it matter?', 'Which groups failed the audit:']],
                                     headers='keys',
                                     tablefmt='pipe', showindex='never', numalign="left") + \
                            oneline

        # mkdown_highlevel += '**The Bias Report has found that the following groups do not have Equal Parity:**\n\n'
        #mkdown_highlevel += get_statpar_text(group_value_df, fairness_measures_depend) + oneline
        mkdown_highlevel += '\n\n[Go to Top](#)' + oneline + '----' + oneline

    if 'Impact Parity' in group_value_df.columns:
        mkdown_highlevel += '\n\n#### Proportional Parity: ' + measures_results_dict['Impact Parity'] + '\n\n'
        raw = {}
        raw['What is it?'] = ['##border##This criteria considers an attribute to have proportional parity if every group is ' \
                              'represented proportionally to their share of the population. For example, if race ' \
                              'with possible values of white, black, other being 50%, 30%, 20% of the population respectively) has ' \
                              'proportional parity, it implies that all three races are represented in the same proportions ' \
                              '(50%, 30%, 20%) in the selected set.']
        raw['When does it matter?'] = ['##border##If your desired outcome is to intervene ' \
                                       'proportionally on people from all races, then you care about this criteria.']
        raw['Which groups failed the audit:'] = '##border##' + get_false_text(group_value_df, 'Impact Parity',
                                                                              fairness_measures_depend)

        dft = pd.DataFrame(raw)
        mkdown_highlevel += tabulate(dft[['What is it?', 'When does it matter?', 'Which groups failed the audit:']],
                                     headers='keys',
                                     tablefmt='pipe', showindex='never', numalign="left") + \
                            oneline

        #mkdown_highlevel += get_impact_text(group_value_df, fairness_measures_depend) + oneline
        mkdown_highlevel += '\n\n[Go to Top](#)' + oneline + '----' + oneline

    if 'FPR Parity' in group_value_df.columns:
        mkdown_highlevel += '\n\n#### False Positive Rate Parity: ' + measures_results_dict['FPR Parity'] + '\n\n'
        # mkdown_highlevel += 'False Positive Rate Parity is concerned with Type I errors (False Positives). In cases ' \
        #                     'of punitive ' \
        #                     'interventions on the selected set' \
        #                     ' it is important to not have disparate Type I errors across groups. Aequitas audits both False ' \
        #                     'Positive Rate (FP/Negative Labels of each group) and False Discovery Rates (FP/Selected ' \
        #                     'Set Size).\n\n'

        raw = {}
        raw['What is it?'] = ['##border##This criteria considers an attribute to have False Positive '
                              'parity if '
                              'every group ' \
                              'has the same False Positive Error Rate. For example, if race has false positive parity, ' \
                              'it implies that all three races have the same False Positive Error Rate.']
        raw['When does it matter?'] = [
            '##border##If your desired outcome is to make false positive '
                                                                       'errors ' \
                                                                       'equally on people from all races, then you care about this criteria. This is important in cases where your intervention is ' \
                                                                       'punitive ' \
                                                                       'and has a risk of adverse outcomes for individuals. Using this criteria allows you to make sure that ' \
                                                                       'you are not making false positive mistakes about any single group disproportionately.']
        raw['Which groups failed the audit:'] = '##border##' + get_false_text(group_value_df, 'FPR Parity',
                                                                              fairness_measures_depend)

        dft = pd.DataFrame(raw)
        mkdown_highlevel += tabulate(
            dft[['What is it?', 'When does it matter?', 'Which groups failed the audit:']],
            headers='keys',
            tablefmt='pipe', showindex='never', numalign="left") + \
                            oneline

        mkdown_highlevel += '\n\n[Go to Top](#)' + oneline
    mkdown_highlevel += oneline + '----' + oneline

    if 'FDR Parity' in group_value_df.columns:
        mkdown_highlevel += oneline + '\n\n#### False Discovery Rate Parity: ' + measures_results_dict['FDR Parity'] + '\n\n'
        raw = {}
        raw['What is it?'] = ['##border##This criteria considers an attribute to have False Discovery Rate parity if every group ' \
                              'has the same False Discovery Error Rate. For example, if race has false discovery parity, ' \
                              'it implies that all three races have the same False Discvery Error Rate.']
        raw['When does it matter?'] = [
            '##border##If your desired outcome is to make false positive '
                                                                        'errors ' \
                                                                        'equally on people from all races, then you care about this criteria. This is important in cases where your intervention is ' \
                                                                        'punitive and can hurt individuals and where you are selecting a very small group for interventions.']
        raw['Which groups failed the audit:'] = '##border##' + get_false_text(group_value_df, 'FDR Parity',
                                                                              fairness_measures_depend)

        dft = pd.DataFrame(raw)
        mkdown_highlevel += tabulate(
            dft[['What is it?', 'When does it matter?', 'Which groups failed the audit:']],
            headers='keys',
            tablefmt='pipe', showindex='never', numalign="left") + \
                            oneline

        mkdown_highlevel += '\n\n[Go to Top](#)' + oneline
    mkdown_highlevel += oneline + '----' + oneline

    if 'FNR Parity' in group_value_df.columns:
        # mkdown_highlevel += 'False Negative Parity is concerned with Type II errors (False Negatives). In cases ' \
        #                     'of assistive or preventive ' \
        #                     'interventions on the selected set' \
        #                     ' it is important to not have disparate Type II errors across groups. Aequitas audits both False ' \
        #                     'Negative Rate (FN/Positive Labels of each group) and False Omission Rates (FN/Not Selected ' \
        #                     'Set Size).\n\n'

        mkdown_highlevel += oneline + '\n\n#### False Negative Rate Parity: ' + measures_results_dict['FNR Parity'] + '\n\n'
        raw = {}
        raw['What is it?'] = ['##border##This criteria considers an attribute to have False Negative parity if every group ' \
                              'has the same False Negative Error Rate. For example, if race has false negative parity, it implies that all three ' \
                              'races ' \
                              'have the same False Negative Error Rate.']
        raw['When does it matter?'] = [
            '##border##If your desired outcome is to make false negative errors equally on ' \
            'people from all races, then you care about this criteria. This is important in cases where your intervention is ' \
            'assistive (providing helpful social services for example) and missing an individual could lead to adverse outcomes for them. Using this criteria allows you to make ' \
            'sure ' \
            'that you’re not missing people from certain groups '
            'disproportionately.']
        raw['Which groups failed the audit:'] = '##border##' + get_false_text(group_value_df, 'FNR Parity',
                                                                              fairness_measures_depend)
        dft = pd.DataFrame(raw)
        mkdown_highlevel += tabulate(dft[['What is it?', 'When does it matter?',
                                          'Which groups failed the audit:']],
                                     headers='keys',
                                     tablefmt='pipe', showindex='never', numalign="left") + \
                            oneline

        mkdown_highlevel += '\n\n[Go to Top](#)' + oneline
    mkdown_highlevel += oneline + '----' + oneline

    if 'FOR Parity' in group_value_df.columns:
        mkdown_highlevel += oneline + '\n\n#### False Omission Rate Parity: ' + measures_results_dict['FOR Parity'] + '\n\n'
        raw = {}
        raw['What is it?'] = ['##border##This criteria considers an attribute to have False Omission Rate parity if every group ' \
                              'has the same False Omission Error Rate. For example, if race has false omission parity, it implies that all three ' \
                              'races ' \
                              'have the same False Omission Error Rate.']
        raw['When does it matter?'] = [
            '##border##If your desired outcome is to make false negative '
                                                                       'errors ' \
                                                                       'equally ' \
                                                                       'on people from all races, then you care about this criteria. This is important in cases where your intervention is ' \
                                                                       'assistive (providing help social services for example) and missing an individual could lead to adverse outcomes for them '\
                                                                       ', and where you are selecting a very small group for interventions. Using this criteria allows you to make ' \
                                                                       'sure ' \
                                                                       'that you’re not missing people from certain groups '
                                                                       'disproportionately.']
        raw['Which groups failed the audit:'] = '##border##' + get_false_text(group_value_df, 'FOR Parity',
                                                                              fairness_measures_depend)

        dft = pd.DataFrame(raw)
        mkdown_highlevel += tabulate(dft[['What is it?', 'When does it matter?',
                                          'Which groups failed the audit:']],
                                     headers='keys',
                                     tablefmt='pipe', showindex='never', numalign="left") + \
                            oneline

        mkdown_highlevel += '\n\n[Go to Top](#)' + oneline
    # mkdown_highlevel += oneline + '----' + oneline

    mkdown_parity = '\n\n### Audit Results: Details by Protected Attributes' + oneline
    # do we want to show this?
    # mkdown_parity += get_highlevel_report(group_attribute_df) + '\n\n'

    mkdown_disparities = '\n\n### Audit Results: Bias Metrics Values' + oneline
    mkdown_group = '\n\n### Audit Results: Group Metrics Values' + oneline
    # setup the group_value_df (colors and stuff)
    group_value_df['group_size_pct'] = group_value_df['group_size'].divide(group_value_df['total_entities'])
    group_value_df = setup_group_value_df(group_value_df, configs.fair_measures_requested,
                                          fairness_measures_depend)
    for attr in configs.attr_cols:
        mkdown_parity += '\n\n#### ' + attr + oneline
        mkdown_disparities += '\n\n#### ' + attr + oneline
        mkdown_group += '\n\n#### ' + attr + oneline
        mkdown_parity += get_parity_group_report(group_value_df, attr, configs.fair_measures_requested, fairness_measures_depend)
        mkdown_disparities += get_disparities_group_report(group_value_df, attr, configs.fair_measures_requested,
                                                           fairness_measures_depend)
        mkdown_group += get_group_group_report(group_value_df, attr, configs.fair_measures_requested,
                                               fairness_measures_depend)
        mkdown_parity += '\n\n[Go to Top](#)' + manylines
        mkdown_disparities += '\n\n[Go to Previous]' + '(#' + '-'.join(attr.lower().split(' ')) + \
                              ')' + '\n\n' + '[Go to Top](#)' + manylines
        mkdown_group += '\n\n[Go to Previous]' + '(#' + '-'.join(attr.lower().split(' ')) + \
                        '-2)' + '\n\n' + '[Go to Top](#)' + manylines

    report = mkdown_highlevel + '----' + mkdown_parity + '----' + mkdown_disparities + '----' + mkdown_group + '----'
    report_html = markdown(report, extras=['tables', 'header-ids'])
    # coloring True/False results
    report_html = report_html.replace('nan', 'Undefined')
    report_html = report_html.replace('>False<', ' style="color:red"><b>Failed</b><')
    report_html = report_html.replace('>True<', ' style="color:green"><b>Passed</b><')

    report_html = report_html.replace('>##red##', ' style="color:red">')
    report_html = report_html.replace('>##green##', ' style="color:green">')

    report_html = report_html.replace('##br##', '<br>')

    report_html = report_html.replace('++span-red-init++', '<span style="color:red"><b>')
    report_html = report_html.replace('++span-red-end++', '</b> </span> ')
    report_html = report_html.replace('++span-green-init++', '<span style="color:green"><b>')
    report_html = report_html.replace('++span-green-end++', '</b> </span> ')

    # report_html = report_html.replace(' failed ', '<span style="color:red"><b> failed </b></span>')
    # report_html = report_html.replace(' passed ', '<span style="color:green"><b> passed </b></span>')

    report_html = report_html.replace('<table>', '<table class="table table-striped" padding=5 >')
    report_html = report_html.replace('<h1 id="the-bias-report">', '<h1 id="the-bias-report" align="center">')
    # report_html = report_html.replace('< thead >\n < tr >', '< thead >\n < tr class="table-info" >')
    #     audit_desc = """<table class="table table-striped" padding=5 >
    # <thead>
    # <tr>
    #   <th align="left">column1</th>
    #   <th align="left">column2</th>
    #   <th align="left">column3</th>
    # </tr>
    # </thead>"""
    #     new_audit_desc = '<table>\n'
    #     report_html = report_html.replace(audit_desc, new_audit_desc)


    ##TOOLTIPS
    # report_html = report_html.replace('Statistical Parity', ' <a href="#" data-toggle="tooltip" title="Hooray!">Equal
    # Parity</a>')

    ##summary table we need a different replace
    report_html = report_html.replace('<td align="left" style="color:red">', '<td align="left" style="color:red; '
                                                                             'padding:10px;">')
    report_html = report_html.replace('<td align="left" style="color:green">', '<td align="left" style="color:green; '
                                                                               'padding:10px;">')
    report_html = report_html.replace('<td align="left">', '<td align="left" style=" padding:5px;">')

    report_html = report_html.replace(';">##border##', '; border: 15px solid white;">')

    ## widths tables
    width1_default = '<th align="left">What is it?</th>'
    width2_default = '<th align="left">Which groups failed the audit:</th>'
    width1_new = '<th align="left" width="35%" >What is it?</th>'
    width3_new = '<th align="left" width="30%" >Which groups failed the audit:</th>'
    report_html = report_html.replace(width1_default, width1_new)
    report_html = report_html.replace(width2_default, width3_new)

    report_html = report_html.replace('##tooltip-start-title##', '<a href="javascript:;" data-toggle="tooltip" title="')
    report_html = report_html.replace('##tooltip-end-title##', ' ">')
    report_html = report_html.replace('##tooltip-end-anchor##', '</a>')

    report_html = report_html.replace('>Statistical Parity<', ' ><a href="javascript:;" data-toggle="tooltip" '
                                                              'title="Fail/Pass test if the Predicted Positive Rate '
                                                              'Disparity of each group is within the range allowed by the fairness '
                                                              'threshold selected.">Equal Parity</a><')



    report_html = report_html.replace('>Impact Parity<', ' ><a href="javascript:;" data-toggle="tooltip" '
                                                         'title="Fail/Pass test if the Predicted Positive Group Rate '
                                                         'Disparity of each group is within the range allowed by the fairness threshold selected.">Proportional Parity</a><')

    report_html = report_html.replace('>FPR Parity<', ' ><a href="javascript:;" data-toggle="tooltip" title="Fail/Pass test '
                                                      'if the False Positive Rate Disparity of each group is within the range '
                                                      'allowed by the fairness threshold selected.">False '
                                                      'Positive '
                                                      'Rate Parity</a><')
    report_html = report_html.replace('>FDR Parity<', '> <a href="javascript:;" data-toggle="tooltip" title="Fail/Pass test '
                                                      'if the False Discovery Rate Disparity of each group is within the range '
                                                      'allowed by the fairness threshold selected.">False Discovery '
                                                      'Rate Parity</a><')
    report_html = report_html.replace('>FNR Parity<', ' ><a href="javascript:;" data-toggle="tooltip" title="Fail/Pass test '
                                                      'if the False Negative Rate Disparity of each group is within the range '
                                                      'allowed by the fairness threshold selected.">False Negative '
                                                      'Rate Parity</a><')
    report_html = report_html.replace('>FOR Parity<', ' ><a href="javascript:;" data-toggle="tooltip" title="Fail/Pass test '
                                                      'if the False Omission Rate Disparity of each group is within the range '
                                                      'allowed by the fairness threshold selected.">False '
                                                      'Omission '
                                                      'Rate Parity</a><')

    report_html = report_html.replace('>Equal Parity<', ' ><a href="javascript:;" data-toggle="tooltip" '
                                                        'title="Fail/Pass test if the Predicted Positive Rate '
                                                        'Disparity of each group is within the range allowed by the fairness '
                                                        'threshold selected.">Equal Parity</a><')


    report_html = report_html.replace('>Proportional Parity<', ' ><a href="javascript:;" data-toggle="tooltip" '
                                                               'title="Fail/Pass test if the Predicted Positive Group Rate '
                                                               'Disparity of each group is within the range allowed by the fairness threshold selected.">Proportional Parity</a><')

    report_html = report_html.replace('>False Positive Rate Parity<', ' ><a href="javascript:;" data-toggle="tooltip" '
                                                                      'title="Fail/Pass test '
                                                                      'if the False Positive Rate Disparity of each group is within the range '
                                                                      'allowed by the fairness threshold selected.">False '
                                                                      'Positive '
                                                                      'Rate Parity</a><')
    report_html = report_html.replace('>False Discovery Rate Parity<', '> <a href="javascript:;" data-toggle="tooltip" '
                                                                       'title="Fail/Pass test '
                                                                       'if the False Discovery Rate Disparity of each group is within the range '
                                                                       'allowed by the fairness threshold selected.">False Discovery '
                                                                       'Rate Parity</a><')
    report_html = report_html.replace('>False Negative Rate Parity<', ' ><a href="javascript:;" data-toggle="tooltip" '
                                                                      'title="Fail/Pass test '
                                                                      'if the False Negative Rate Disparity of each group is within the range '
                                                                      'allowed by the fairness threshold selected.">False Negative '
                                                                      'Rate Parity</a><')
    report_html = report_html.replace('>False Omission Rate Parity<', ' ><a href="javascript:;" data-toggle="tooltip" '
                                                                      'title="Fail/Pass test '
                                                                      'if the False Omission Rate Disparity of each group is within the range '
                                                                      'allowed by the fairness threshold selected.">False '
                                                                      'Omission '
                                                                      'Rate Parity</a><')

    report_html = report_html.replace('>PPREV Disparity<',
                                      ' ><a href="javascript:;" data-toggle="tooltip" title="The Predicted Positive Group Ratio of each group divided by the same metric value of the reference group. '
                                      '.">Predicted Positive '
                                      'Group Rate Disparity'
                                      '</a><')

    report_html = report_html.replace('>PPR Disparity<', ' ><a href="javascript:;" data-toggle="tooltip" title="The Predicted '
                                                         'Positive Rate of each group divided by the same metric value of the reference group.">Predicted '
                                                         'Positive '
                                                         'Rate Disparity</a><')

    report_html = report_html.replace('>FPR Disparity<', ' ><a href="javascript:;" data-toggle="tooltip" title="The False '
                                                         'Positive Rate of each group divided by the same metric value of the '
                                                         'reference group.">False '
                                                         'Positive '
                                                         'Rate Disparity</a><')
    report_html = report_html.replace('>FDR Disparity<', '> <a href="javascript:;" data-toggle="tooltip" title="The False '
                                                         'Discovery Rate of each group divided by the same metric value of the '
                                                         'reference group.">False Discovery '
                                                         'Rate Disparity</a><')
    report_html = report_html.replace('>FNR Disparity<', ' ><a href="javascript:;" data-toggle="tooltip" title="The False '
                                                         'Negative Rate of each group divided by the same metric value of the '
                                                         'reference group.">False Negative '
                                                         'Rate Disparity</a><')
    report_html = report_html.replace('>FOR Disparity<', ' ><a href="javascript:;" data-toggle="tooltip" title="The False '
                                                         'Omission Rate of each group divided by the same '
                                                         'metric value of the reference group.">False '
                                                         'Omission '
                                                         'Rate Disparity</a><')

    report_html = report_html.replace('>PPREV<',
                                      ' ><a href="javascript:;" data-toggle="tooltip" title="Number of elements of the '
                                      'group selected divided by the size of the group.">Predicted Positive '
                                      'Group Rate'
                                      '</a><')

    report_html = report_html.replace('>PPR<', ' ><a href="javascript:;" data-toggle="tooltip" title="Number of elements of the '
                                               'group selected divided by the total size of the interventation set.">Predicted '
                                               'Positive '
                                               'Rate</a><')

    report_html = report_html.replace('>FPR<',
                                      ' ><a href="javascript:;" data-toggle="tooltip" title="Number of false positives of '
                                      'the group divided by number of labeled negatives within the group.">False '
                                      'Positive '
                                      'Rate</a><')
    report_html = report_html.replace('>FDR<',
                                      '> <a href="javascript:;" data-toggle="tooltip" title="Number of false positives of '
                                      'the group divided by the size of the intervention set of the group (predicted '
                                      'positives within group).">False Discovery '
                                      'Rate</a><')
    report_html = report_html.replace('>FNR<',
                                      ' ><a href="javascript:;" data-toggle="tooltip" title="Number of false negative of '
                                      'the group divided by number of labeled positives in the group">False Negative '
                                      'Rate</a><')
    report_html = report_html.replace('>FOR<',
                                      ' ><a href="javascript:;" data-toggle="tooltip" title="Number of false negatives of '
                                      'the group divided by the size of the non-selected set of the group (predicted '
                                      'negatives within group).">False '
                                      'Omission '
                                      'Rate</a><')

    audit_desc = """<table class="table table-striped" padding=5 >
<thead>
<tr>
  <th align="left">column1</th>
  <th align="left">column2</th>
</tr>
</thead>"""
    new_audit_desc = '<table>\n'
    report_html = report_html.replace(audit_desc, new_audit_desc)

    audit_desc = """<table class="table table-striped" padding=5 >
<thead>
<tr>
  <th align="left">column1</th>
  <th align="left">column2</th>
  <th align="left">column3</th>
</tr>
</thead>"""
    new_audit_desc = '<table>\n'
    report_html = report_html.replace(audit_desc, new_audit_desc)
    # change color of headers
    report_html = report_html.replace('<h1', '<h1 style="color:#b37d4e;"')
    report_html = report_html.replace('<h3', '<h3 style="color:#b37d4e;"')
    report_html = report_html.replace('<h3', '<h3 style="color:#b37d4e;"')
    report_html = report_html.replace('<h4', '<h4 style="color:#286da8;"')
    report_html = report_html.replace('<strong ><a', '<a')
    report_html = report_html.replace('</a></strong>', '</a>')
    report_html = report_html.replace('<strong> <a', '<a')

    return report_html



import logging

import pandas as pd

logging.getLogger(__name__)

__author__ = "Rayid Ghani, Pedro Saleiro <saleiro@uchicago.edu>, Benedict Kuester, Loren Hinkson"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."


class Group(object):
    """
    """
    def __init__(self):
        """
        """
        self.label_neg_count = lambda label_col: lambda x: \
            (x[label_col] == 0).sum()
        self.label_pos_count = lambda label_col: lambda x: \
            (x[label_col] == 1).sum()
        self.group_functions = self._get_group_functions()

    @staticmethod
    def _get_group_functions():
        """
        Helper function to accumulate lambda functions used in bias metrics
        calculations.
        """

        divide = lambda x, y: x / y if y != 0 else pd.np.nan

        predicted_pos_count = lambda rank_col, label_col, thres, k: lambda x: \
            (x[rank_col] <= thres).sum()

        predicted_neg_count = lambda rank_col, label_col, thres, k: lambda x: \
            (x[rank_col] > thres).sum()

        predicted_pos_ratio_k = lambda rank_col, label_col, thres, k: lambda x: \
            divide((x[rank_col] <= thres).sum(), k + 0.0)

        predicted_pos_ratio_g = lambda rank_col, label_col, thres, k: lambda x: \
            divide((x[rank_col] <= thres).sum(), len(x) + 0.0)

        false_neg_count = lambda rank_col, label_col, thres, k: lambda x: \
            ((x[rank_col] > thres) & (x[label_col] == 1)).sum()

        false_pos_count = lambda rank_col, label_col, thres, k: lambda x: \
            ((x[rank_col] <= thres) & (x[label_col] == 0)).sum()

        true_neg_count = lambda rank_col, label_col, thres, k: lambda x: \
            ((x[rank_col] > thres) & (x[label_col] == 0)).sum()

        true_pos_count = lambda rank_col, label_col, thres, k: lambda x: \
            ((x[rank_col] <= thres) & (x[label_col] == 1)).sum()

        fpr = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] <= thres) & (x[label_col] == 0)).sum(),
                   (x[label_col] == 0).sum().astype(
                       float))

        tnr = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] > thres) & (x[label_col] == 0)).sum(), (x[label_col] ==
                                                                         0).sum().astype(
                float))

        fnr = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] > thres) & (x[label_col] == 1)).sum(),
                   (x[label_col] == 1).sum().astype(
                       float))

        tpr = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] <= thres) & (x[label_col] == 1)).sum(), (x[label_col] ==
                                                                          1).sum().astype(
                float))

        fomr = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] > thres) & (x[label_col] == 1)).sum(), (x[rank_col] >
                                                                         thres).sum(
            ).astype(float))

        npv = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] > thres) & (x[label_col] == 0)).sum(),
                   (x[rank_col] > thres).sum().astype(
                       float))

        precision = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] <= thres) & (x[label_col] == 1)).sum(), (x[rank_col] <=
                                                                          thres).sum(
            ).astype(float))

        fdr = lambda rank_col, label_col, thres, k: lambda x: \
            divide(((x[rank_col] <= thres) & (x[label_col] == 0)).sum(), (x[rank_col] <=
                                                                          thres).sum(
            ).astype(float))

        group_functions = {'tpr': tpr,
                           'tnr': tnr,
                           'for': fomr,
                           'fdr': fdr,
                           'fpr': fpr,
                           'fnr': fnr,
                           'npv': npv,
                           'precision': precision,
                           'pp': predicted_pos_count,
                           'pn': predicted_neg_count,
                           'ppr': predicted_pos_ratio_k,
                           'pprev': predicted_pos_ratio_g,
                           'fp': false_pos_count,
                           'fn': false_neg_count,
                           'tn': true_neg_count,
                           'tp': true_pos_count}

        return group_functions

    def get_crosstabs(self, df, score_thresholds=None, model_id=1, attr_cols=None):
        """
        Creates univariate groups and calculates group metrics.

        :param df: a dataframe containing the following required columns [score,  label_value].
        :param score_thresholds: dictionary { 'rank_abs':[] , 'rank_pct':[], 'score':[] }
        :param model_id: the model ID on which to subset the df.
        :param attr_cols: optional, list of names of columns corresponding to
            group attributes (i.e., gender, age category, race, etc.).

        :return: A dataframe of group score, label, and error statistics and absolute bias metric values grouped by unique attribute values
        """
        if not attr_cols:
            non_attr_cols = ['id', 'model_id', 'entity_id', 'score', 'label_value', 'rank_abs', 'rank_pct']
            attr_cols = df.columns[~df.columns.isin(non_attr_cols)]  # index of the columns that are
        # check if all attr_cols exist in df
        check = [col in df.columns for col in attr_cols]
        if False in check:
            # todo: create separate check method that raises exception...
            logging.error('get_crosstabs: not all attribute columns provided exist in input dataframe!')
            # exit(1)
        # check if all columns are strings:
        non_string_cols = df.columns[(df.dtypes != object) & (df.dtypes != str) & (df.columns.isin(attr_cols))]
        if non_string_cols.empty is False:
            logging.error('get_crosstabs: input df was not preprocessed. There are non-string cols within attr_cols!')
            # exit(1)

        # if no score_thresholds are provided, we assume that rank_abs=number of 1s in the score column
        count_ones = None  # it also serves as flag to set parameter to 'binary'

        if not score_thresholds:
            df['score'] = df['score'].astype(float)
            count_ones = df['score'].value_counts().get(1.0, 0)
            score_thresholds = {'rank_abs': [count_ones]}

        print('model_id, score_thresholds', model_id, score_thresholds)
        df = df.sort_values('score', ascending=False)
        df['rank_abs'] = range(1, len(df) + 1)
        df['rank_pct'] = df['rank_abs'] / len(df)
        dfs = []
        prior_dfs = []
        # calculate the bias for these columns
        # not default(non_attr_cols), therefore represent the group variables!
        logging.info('getcrosstabs: attribute columns to perform crosstabs:' + ','.join(attr_cols))
        # for each group variable do
        for col in attr_cols:
            # find the priors_df
            col_group = df.fillna({col: 'pd.np.nan'}).groupby(col)
            counts = col_group.size()
            print('COUNTS:::', counts)
            # distinct entities within group value
            this_prior_df = pd.DataFrame({
                'model_id': [model_id] * len(counts),
                'attribute_name': [col] * len(counts),
                'attribute_value': counts.index.values,
                'group_label_pos': col_group.apply(self.label_pos_count(
                    'label_value')).values,
                'group_label_neg': col_group.apply(self.label_neg_count(
                    'label_value')).values,
                'group_size': counts.values,
                'total_entities': [len(df)] * len(counts)
            })
            this_prior_df['prev'] = this_prior_df['group_label_pos'] / this_prior_df['group_size']
            # for each model_id and as_of_date the priors_df has length
            # attribute_names * attribute_values
            prior_dfs.append(this_prior_df)

            # we calculate the bias for two different types of score_thresholds
            # units (percentage ranks and absolute ranks)
            # YAML ex: thresholds:
            #              rank_abs: [300]
            #              rank_pct: [1.0, 5.0, 10.0]
            for thres_unit, thres_values in score_thresholds.items():

                for thres_val in thres_values:
                    flag = 0

                    # To discuss with Pedro: believe this might be the reason
                    # for cutoff error - if numbers are cumulative, per
                    # line 149 and line 150, why taking sum for k vs. max?
                    k = (df[thres_unit] <= thres_val).sum()

                    # denote threshold as binary if numeric count_ones value
                    # donate as [rank value]_abs or [rank_value]_pct otherwise
                    score_threshold = 'binary 0/1' if count_ones != None else str(thres_val) + '_' + thres_unit[-3:]
                    for name, func in self.group_functions.items():
                        func = func(thres_unit, 'label_value', thres_val, k)
                        feat_bias = col_group.apply(func)
                        metrics_df = pd.DataFrame({
                            'model_id': [model_id] * len(feat_bias),
                            'score_threshold': [score_threshold] * len(feat_bias),
                            'k': [k] * len(feat_bias),
                            'attribute_name': [col] * len(feat_bias),
                            'attribute_value': feat_bias.index.values,
                            name: feat_bias.values
                        })
                        if flag == 0:
                            this_group_df = metrics_df
                            flag = 1
                        else:
                            this_group_df = this_group_df.merge(metrics_df)
                        # print(this_group_df.head(1))
                    dfs.append(this_group_df)
        # precision@	25_abs
        groups_df = pd.concat(dfs, ignore_index=True)
        priors_df = pd.concat(prior_dfs, ignore_index=True)
        groups_df = groups_df.merge(priors_df, on=['model_id', 'attribute_name',
                                                   'attribute_value'])
        return groups_df, attr_cols

    def list_absolute_metrics(self, df):
        """
        View list of all calculated absolute bias metrics in df
        """
        return df.columns.intersection(['fpr', 'fnr', 'tpr', 'tnr', 'for',
                                           'fdr', 'npv', 'precision', 'ppr',
                                           'pprev', 'prev'
                                        ]).tolist()



import logging
from datetime import datetime
from os import path
from sys import exit

import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from xhtml2pdf import pisa

logging.getLogger(__name__)

__author__ = "Rayid Ghani, Pedro Saleiro <saleiro@uchicago.edu>"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."


def get_db_data(engine, input_query):
    """

    :param engine:
    :param input_query:
    :return:
    """
    logging.info('querying db...')
    try:
        df = pd.read_sql(input_query, engine)
    except SQLAlchemyError as e:
        logging.error('PG: could not get the resulting table. Please check your input_query. ' + e)
        exit(1)
    return df


def get_csv_data(input_file):
    """

    :param input_file:
    :return:
    """
    logging.info('loading csv data...')
    try:
        df = pd.read_csv(input_file)
    except IOError:
        logging.error('run_csv: could not load csv provided as input.')
        exit(1)
    return df


def push_todb(engine, output_schema, create_tables, output_df):
    """

    :param engine:
    :param output_df:
    :return:
    """
    logging.info('pushing to db aequitas table...')
    try:
        output_df.set_index(['model_id', 'attribute_name']).to_sql(
            schema=output_schema,
            name='aequitas_group',
            con=engine,
            if_exists=create_tables)
    except SQLAlchemyError as e:
        logging.error('push_db_data: Could not push results to the target database.' + e)


def push_tocsv(input_file, output_df):
    """

    :param input_file:
    :param output_folder:
    :param output_df:
    :return:
    """
    logging.info('pushing to csv...')
    datestr = datetime.now().strftime("%Y%m%d-%H%M%S")
    ipath, input_name = path.split(input_file)
    # outpath = output_folder + input_name[:input_file.find('.')] + '_aequitas_' + datestr + '.csv'
    # for now we are saving the csv in same folder as the input file
    outpath = ipath + input_name[:input_file.find('.')] + '_aequitas_' + datestr + '.csv'

    try:
        output_df.to_csv(outpath, encoding='utf-8', index=False)
    except IOError:
        logging.info('push_csv_data: Could not push results to a csv file.')


def push_topdf(input_file, report):
    datestr = datetime.now().strftime("%Y%m%d-%H%M%S")
    ipath, input_name = path.split(input_file)
    # outpath = output_folder + input_name[:input_file.find('.')] + '_aequitas_' + datestr + '.csv'
    # for now we are saving the csv in same folder as the input file
    try:
        outpath = ipath + input_name[:input_file.find('.')] + '_aequitas_' + datestr + '.pdf'
        logging.info('Saving bias report to pdf on ' + outpath)

        result_pdf = open(outpath, 'w+b')
        pisaStatus = pisa.CreatePDF(report, dest=result_pdf)
        result_pdf.close()
    except IOError:
        logging.info('Failed! Not able to save Aequitas Report pdf on ' + outpath)
    return


def get_engine(configs):
    """

    :param configs:
    :return: Returns a SQLAlchemy pg connnection.
    """
    try:
        db_access = configs.db['db_credentials']
    except KeyError:
        logging.error('KeyError in configuration file.')
        exit(1)
    try:
        engine = create_engine('postgresql://{user}:{pw}@{host}:{port}/{db}'.format(
            user=db_access['user'],
            pw=db_access['password'],
            host=db_access['host'],
            port=db_access['port'],
            db=db_access['database']))
    except SQLAlchemyError:
        logging.error('PG: could not create pg engine!')
        exit(1)
    return engine



import logging
import math
import numpy as np
import collections
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.colors
import matplotlib.cm


from aequitas import squarify_flipped as sf

logging.getLogger(__name__)

__author__ = "Pedro Saleiro <saleiro@uchicago.edu>, Loren Hinkson"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."



# module-level function
def assemble_ref_groups(disparities_table, ref_group_flag='_ref_group_value',
                         specific_measures=None, label_score_ref=None):
    """
    Creates a dictionary of reference groups for each metric in a data_table.

   :param disparities_table: a disparity table. Output of bias.get_disparity or
        fairness.get_fairness functions
    :param ref_group_flag: string indicating column indicates reference group
        flag value. Default is '_ref_group_value'.
    :param specific_measures: Limits reference dictionary to only specified
        metrics in a data table. Default is None.
    :param label_score_ref: Defines a metric, ex: 'fpr' (false positive rate)
        from which to mimic reference group for label_value and score. Used for
        statistical significance calculations in Bias() class. Default is None.

    :return: A dictionary
    """
    ref_groups = {}
    ref_group_cols = \
        list(disparities_table.columns[disparities_table.columns.str.contains(
            ref_group_flag)])

    if specific_measures:
        ref_group_cols = \
            [measure + ref_group_flag for measure in specific_measures if
             measure + ref_group_flag in ref_group_cols]

    attributes = list(disparities_table.attribute_name.unique())
    for attribute in attributes:
        attr_table = \
            disparities_table.loc[disparities_table['attribute_name'] == attribute]
        attr_refs = {}
        for col in ref_group_cols:
            metric_key = "".join(col.split(ref_group_flag))
            attr_refs[metric_key] = \
                attr_table.loc[attr_table['attribute_name'] == attribute, col].min()
        if label_score_ref:
            if label_score_ref + ref_group_flag in ref_group_cols:
                attr_refs['label_value'] = attr_refs[label_score_ref]
                attr_refs['score'] = attr_refs[label_score_ref]
            else:
                raise ValueError("The specified reference measure for label"
                                 " value and score is not included in the "
                                 "data frame.")

        ref_groups[attribute] = attr_refs

    return ref_groups


# Plot() class
class Plot(object):
    """
    Plotting object allows for visualization of absolute group bias metrics and
    relative disparities calculated by Aequitas Group(), Bias(), and Fairness()
    class instances.
    """
    default_absolute_metrics = ('pprev', 'ppr', 'fdr', 'for', 'fpr', 'fnr')
    default_disparities = ('pprev_disparity', 'ppr_disparity',
                           'fdr_disparity', 'for_disparity',
                           'fpr_disparity', 'fnr_disparity')

    # Define mapping for conditional coloring based on fairness
    # determinations
    _metric_parity_mapping = {
        'ppr_disparity': 'Statistical Parity',
        'pprev_disparity': 'Impact Parity',
        'precision_disparity': 'Precision Parity',
        'fdr_disparity': 'FDR Parity',
        'for_disparity': 'FOR Parity',
        'fpr_disparity': 'FPR Parity',
        'fnr_disparity': 'FNR Parity',
        'tpr_disparity': 'TPR Parity',
        'tnr_disparity': 'TNR Parity',
        'npv_disparity': 'NPV Parity',
        'ppr': 'Statistical Parity',
        'pprev': 'Impact Parity',
        'precision': 'Precision Parity',
        'fdr': 'FDR Parity',
        'for': 'FOR Parity',
        'fpr': 'FPR Parity',
        'fnr': 'FNR Parity',
        'tpr': 'TPR Parity',
        'tnr': 'TNR Parity',
        'npv': 'NPV Parity'
    }

    _significance_disparity_mapping = {
        'ppr_disparity': 'ppr_significance',
        'pprev_disparity': 'pprev_significance',
        'precision_disparity': 'precision_significance',
        'fdr_disparity': 'fdr_significance',
        'for_disparity': 'fnr_significance',
        'fpr_disparity': 'fpr_significance',
        'fnr_disparity': 'fnr_significance',
        'tpr_disparity': 'tpr_significance',
        'tnr_disparity': 'tnr_significance',
        'npv_disparity': 'npv_significance'
    }

    def __init__(self, key_metrics=default_absolute_metrics,
                 key_disparities=default_disparities):
        """
        :param key_metrics: Set default absolute group metrics for all subplots
        :param key_disparities: Set default disparity metrics for all subplots
        """
        self.key_metrics = key_metrics
        self.key_disparities = key_disparities

    @staticmethod
    def _nearest_quartile(x):
        '''
        Return nearest quartile for given value x.
        '''
        rounded = round(x * 4) / 4
        if rounded > x:
            return rounded
        else:
            return rounded + 1 / 4

    @staticmethod
    def _check_brightness(rgb_tuple):
        '''
        Determine the brightness of background color in a plot.

        Adapted from https://trendct.org/2016/01/22/how-to-choose-a-label-color-to-contrast-with-background/
        '''
        r, g, b = rgb_tuple
        return (r * 299 + g * 587 + b * 114) / 1000

    @classmethod
    def _brightness_threshold(cls, rgb_tuple, min_brightness, light_color,
                             dark_color='black'):
        '''
        Determine ideal plot label color (light or dark) based on brightness of
        background color based on a given brightness threshold.

        Adapted from https://trendct.org/2016/01/22/how-to-choose-a-label-color-to-contrast-with-background/
        '''
        if cls._check_brightness(rgb_tuple) > min_brightness:
            return dark_color

        return light_color

    @staticmethod
    def _truncate_colormap(orig_cmap, min_value=0.0, max_value=1.0, num_colors=100):
        '''
        Use only part of a colormap (min_value to max_value) across a given number
        of partitions.

        :param orig_cmap: an existing Matplotlib colormap.
        :param min_value: desired minimum value (0.0 to 1.0) for truncated
            colormap. Default is 0.0.
        :param max_value: desired maximum value (0.0 to 1.0) for truncated
            colormap. Default is 1.0.
        :param num_colors: number of colors to spread colormap gradient across
            before truncating. Default is 100.
        :return: Truncated color map

        Attribution: Adapted from: https://stackoverflow.com/questions/
        18926031/how-to-extract-a-subset-of-a-colormap-as-a-new-colormap-in-matplotlib
        '''
        cmap = plt.get_cmap(orig_cmap)
        new_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
            'trunc({n},{a:.2f},{b: .2f})'.format(n=cmap.name, a=min_value, b=max_value),
            cmap(np.linspace(min_value, max_value, num_colors)))
        return new_cmap


    @classmethod
    def _locate_ref_group_indices(cls, disparities_table, attribute_name, group_metric,
                                 ref_group_flag='_ref_group_value', model_id=1):
        """
        Finds relative index (row) of reference group value for a given metric.

        :param disparities_table: a disparity table. Output of bias.get_disparity or
            fairness.get_fairness functions.
        :param attribute_name: the attribute to plot metric against. Must be a column
            in the disparities_table.
        :param group_metric: the metric to plot. Must be a column in the
            disparities_table.
        :param ref_group_flag: string indicating column indicates reference group
            flag value. Default is '_ref_group_value'.
        :param model_id: model ID number. Default is 1.

        :return: Integer indicating relative index of reference group value row.
        """
        # get absolute metric name from passed group metric (vs. a disparity name)
        abs_metric = "".join(group_metric.split('_disparity'))

        all_ref_groups = assemble_ref_groups(disparities_table, ref_group_flag)
        ref_group_name = all_ref_groups[attribute_name][abs_metric]

        # get index for row associated with reference group for that model
        ind = list(disparities_table.loc[(disparities_table['attribute_name'] == attribute_name) &
                                         (disparities_table['attribute_value'] == ref_group_name) &
                                         (disparities_table['model_id'] == model_id)].index)

        # there should only ever be one item in list, but JIC, select first
        idx = ind[0]

        relative_ind = disparities_table.index.get_loc(idx)
        return relative_ind, ref_group_name

    def plot_group_metric(self, group_table, group_metric, ax=None, ax_lim=None,
                          title=True, label_dict=None, min_group_size = None):
        """
        Plot a single group metric across all attribute groups.

        :param group_table: group table. Output of of group.get_crosstabs() or
            bias.get_disparity functions.
        :param group_metric: the metric to plot. Must be a column in the group_table.
        :param ax: a matplotlib Axis. If not passed, a new figure will be created.
        :param title: whether to include a title in visualizations. Default is True.
        :param label_dict: optional, dictionary of replacement labels for data.
            Default is None.
        :param min_group_size: minimum size for groups to include in visualization
            (as a proportion of total sample)

        :return: A Matplotlib axis
        """
        if group_metric not in group_table.columns:
            raise ValueError(f"Specified disparity metric '{group_metric}' not "
                             f"in 'group_table'.")

        if group_table[group_metric].isnull().any():
            raise ValueError(f"Cannot plot {group_metric}, has NaN values.")

        if ax is None:
            (_fig, ax) = plt.subplots(figsize=(10, 5))

        height_of_bar = 1
        attribute_names = group_table.attribute_name.unique()
        tick_indices = []
        next_bar_height = 0

        if min_group_size:
            if min_group_size > (group_table.group_size.max() / group_table.group_size.sum()):
                raise ValueError(f"'min_group_size' proportion specified: '{min_group_size}' "
                                f"is larger than all groups in sample.")

            min_size = min_group_size * group_table.group_size.sum()
            group_table = group_table.loc[group_table['group_size'] >= min_size]

        label_position_values = collections.deque(group_table[group_metric].values)

        lighter_coppers = self._truncate_colormap('copper_r', min_value=0,
                                                   max_value=0.65)

        norm = matplotlib.colors.Normalize(vmin=group_table['group_size'].min(),
                                vmax=group_table['group_size'].max())
        mapping = matplotlib.cm.ScalarMappable(norm=norm, cmap=lighter_coppers)

        # Lock absolute value metric plot x-axis to (0, 1)
        if not ax_lim:
            ax_lim = 1
        ax.set_xlim(0, ax_lim)

        for attribute_name in attribute_names:

            attribute_data = group_table.loc[
                (group_table['attribute_name'] == attribute_name)]

            values = attribute_data[group_metric].values
            grp_sizes = attribute_data['group_size'].values

            attribute_indices = np.arange(next_bar_height,
                                          next_bar_height + attribute_data.shape[0],
                                          step=height_of_bar)
            attribute_tick_location = float((min(attribute_indices) + max(attribute_indices) + height_of_bar)) / 2

            h_attribute = ax.barh(attribute_indices,
                                  width=values,
                                  # label=list(attribute_data['attribute_value'].values),
                                  align='edge', edgecolor='grey')

            label_colors = []
            min_brightness = 0.55

            for bar, g_size in zip(h_attribute, grp_sizes):
                my_col = mapping.to_rgba(g_size)
                bar.set_color(my_col)
                label_colors.append(self._brightness_threshold(
                    rgb_tuple=my_col[:3], min_brightness=min_brightness,
                    light_color=(1, 1, 1, 1)))

            if label_dict:
                labels = [label_dict.get(label, label) for label in
                          attribute_data['attribute_value'].values]
            else:
                labels = attribute_data['attribute_value'].values

            for y, label, value, text_color, g_size in zip(attribute_indices, labels,
                                                   values, label_colors,
                                                   grp_sizes):
                next_position = label_position_values.popleft()
                group_label = f"{label} ({g_size:,})"

                if ax_lim < 3:
                    CHAR_PLACEHOLDER = 0.03
                else:
                    CHAR_PLACEHOLDER = 0.25

                label_length = len(label) * CHAR_PLACEHOLDER
                max_val_length = 7 * CHAR_PLACEHOLDER
                indent_length = ax_lim * 0.025

                # bar long enough for label, enough space after bar for value
                if ((indent_length + label_length) < (next_position - indent_length)) and (
                        (next_position + indent_length + max_val_length) < (
                        ax_lim - indent_length)):

                    ax.text(next_position + indent_length, y + float(height_of_bar) / 2,
                            f"{value:.2f}", fontsize=12, verticalalignment='top')
                    ax.text(indent_length, y + float(height_of_bar) / 2,
                            group_label, fontsize=11, verticalalignment='top',
                            color=text_color)

                # case when bar too long for labels after bar, print all text in bar
                elif (next_position + indent_length + max_val_length) > (
                        ax_lim - indent_length):

                    ax.text(indent_length, y + float(height_of_bar) / 2,
                            f"{group_label}, {value:.2f}", fontsize=11,
                            verticalalignment='top', color=text_color)

                # case when bar too small for labels inside bar, print after bar
                else:
                    ax.text(next_position + indent_length, y + float(
                        height_of_bar) / 2,
                            f"{group_label}, {value:.2f}", fontsize=12,
                            verticalalignment='top')

            tick_indices.append((attribute_name, attribute_tick_location))
            next_bar_height = max(attribute_indices) + 2 * height_of_bar

        ax.yaxis.set_ticks(list(map(lambda x: x[1], tick_indices)))
        ax.yaxis.set_ticklabels(list(map(lambda x: x[0], tick_indices)), fontsize=14)
        ax.set_axisbelow(True)
        ax.xaxis.grid(color='lightgray', which='major',linestyle='dashed')
        ax.set_xlabel("Absolute Metric Magnitude")

        if title:
            ax.set_title(f"{group_metric.upper()}", fontsize=20)

        return ax

    def plot_disparity(self, disparity_table, group_metric, attribute_name,
                       color_mapping=None, model_id=1, ax=None, fig=None,
                       label_dict=None, title=True,
                       highlight_fairness=False, min_group_size=None,
                       significance_alpha=0.05):
        """
        Create treemap based on a single bias disparity metric across attribute
        groups.

        Adapted from https://plot.ly/python/treemaps/,
        https://gist.github.com/gVallverdu/0b446d0061a785c808dbe79262a37eea,
        and https://fcpython.com/visualisation/python-treemaps-squarify-matplotlib

        :param disparity_table: a disparity table. Output of bias.get_disparity or
            fairness.get_fairness function.
        :param group_metric: the metric to plot. Must be a column in the
            disparity_table.
        :param attribute_name: which attribute to plot group_metric across.
        :param color_mapping: matplotlib colormapping for treemap value boxes.
        :param model_id: model ID number. Default is 1.
        :param ax: a matplotlib Axis. If not passed, a new figure will be created.
        :param fig: a matplotlib Figure. If not passed, a new figure will be created.
        :param label_dict: optional, dictionary of replacement labels for data.
            Default is None.
        :param title: whether to include a title in visualizations. Default is True.
        :param highlight_fairness: whether to highlight treemaps by disparity
            magnitude, or by related fairness determination.
        :param min_group_size: minimum proportion of total group size (all data)
            a population group must meet in order to be included in bias metric
            visualization
        :param significance_alpha: statistical significance level. Used to
            determine visual representation of significance (number of
            asterisks on treemap).

        :return: A Matplotlib axis
        """
        # Use matplotlib to truncate colormap, scale metric values
        # between the min and max, then assign colors to individual values

        table_columns = set(disparity_table.columns)
        if group_metric not in table_columns:
           raise ValueError(f"Specified disparity metric {group_metric} not in 'disparity_table'.")

        attribute_table = \
            disparity_table.loc[disparity_table['attribute_name'] == attribute_name]

        # sort by group size, as box size is indicative of group size
        sorted_df = attribute_table.sort_values('group_size', ascending=False)

        x = 0.
        y = 0.
        width = 100.
        height = 100.

        ref_group_rel_idx, ref_group_name = \
            self._locate_ref_group_indices(disparities_table=sorted_df,
                                            attribute_name=attribute_name,
                                            group_metric=group_metric)

        if min_group_size:
            if min_group_size > (disparity_table.group_size.max() /
                            disparity_table.group_size.sum()):
                raise ValueError(f"'min_group_size' proportion specified: '{min_group_size}' "
                                f"is larger than all groups in sample.")

            min_size = min_group_size * disparity_table.group_size.sum()

            # raise warning if minimum group size specified would exclude
            # reference group
            if any(sorted_df.loc[(sorted_df['attribute_value']==ref_group_name),
                                 ['group_size']].values < min_size):
                logging.warning(
                    f"Reference group size is smaller than 'min_group_size' proportion "
                    f"specified: '{min_group_size}'. Reference group '{ref_group_name}' "
                    f"was not excluded.")

            sorted_df = \
                    sorted_df.loc[(sorted_df['group_size'] >= min_size) |
                                  (sorted_df['attribute_value'] == ref_group_name)]

        # select group size as values for size of boxes
        values = sorted_df.loc[:, 'group_size']

        # get new index for ref group
        ref_group_rel_idx, _ = \
            self._locate_ref_group_indices(disparities_table=sorted_df,
                                            attribute_name=attribute_name,
                                            group_metric=group_metric)

        # labels for squares in tree map:
        # label should always be disparity value (but boxes visualized should be
        # always be the metric absolute value capped between 0.1x ref group and
        # 10x ref group)
        if group_metric + '_disparity' not in attribute_table.columns:
            related_disparity = group_metric

        else:
            related_disparity = group_metric + '_disparity'


        if highlight_fairness:
            if not len(table_columns.intersection(self._metric_parity_mapping.values())) > 1:
                raise ValueError("Data table must include at least one fairness "
                                 "determination to visualize metric parity.")

            # apply red for "False" fairness determinations and green for "True"
            # determinations
            cb_green = '#1b7837'
            cb_red = '#a50026'

            parity = self._metric_parity_mapping[group_metric]
            if (parity not in table_columns):
                raise ValueError(
                    f"Related fairness determination for {group_metric} must be "
                    f"included in data table to color visualization based on "
                    f"metric fairness.")
            clrs = [cb_green if val else cb_red for val in sorted_df[parity]]

        else:
            aq_palette = sns.diverging_palette(225, 35, sep=10, as_cmap=True)

            if not color_mapping:
                norm = matplotlib.colors.Normalize(vmin=0, vmax=2)
                color_mapping = matplotlib.cm.ScalarMappable(norm=norm, cmap=aq_palette)

            clrs = \
                [color_mapping.to_rgba(val) for val in sorted_df[related_disparity]]

        # color reference group grey
        clrs[ref_group_rel_idx] = '#D3D3D3'

        compare_value = values.iloc[ref_group_rel_idx]

        scaled_values = [(0.1 * compare_value) if val < (0.1 * compare_value) else
                         (10 * compare_value) if val >= (10 * compare_value) else
                         val for val in values]

        label_values = \
            ["(Ref)" if attr_val == ref_group_name else
             f"{disp:.2f}" for attr_val, disp in
             zip(sorted_df['attribute_value'], sorted_df[related_disparity]) ]

        if label_dict:
            labels = \
                [label_dict.get(label, label) for label in sorted_df['attribute_value']]
        else:
            labels = sorted_df['attribute_value'].values

        # if df includes significance columns, add stars to indicate significance
        if sorted_df.columns[
            sorted_df.columns.str.contains('_significance')].value_counts().sum() > 0:
        # unmasked significance
        # find indices where related significance have smaller value than significance_alpha
            if np.issubdtype(
                    sorted_df[
                        self._significance_disparity_mapping[related_disparity]].dtype,
                    np.number):
                to_star = sorted_df.loc[
                    sorted_df[
                        self._significance_disparity_mapping[related_disparity]] < significance_alpha].index.tolist()


            # masked significance
            # find indices where attr values have True value for each of those two columns,
            else:
                to_star = sorted_df.loc[
                    sorted_df[
                        self._significance_disparity_mapping[related_disparity]] > 0].index.tolist()


            # add stars to label value where significant
            for idx in to_star:
                # convert idx location to relative index in sorted df and label_values list
                idx_adj = sorted_df.index.get_loc(idx)

                # star significant disparities in visualizations based on significance level
                if 0.10 >= significance_alpha > 0.05:
                    significance_stars = '*'
                elif 0.05 >= significance_alpha > 0.01:
                    significance_stars = '**'
                elif significance_alpha <= 0.01:
                    significance_stars = '***'
                else:
                    significance_stars = ''
                label_values[idx_adj] = label_values[idx_adj] + significance_stars


        normed = sf.normalize_sizes(scaled_values, width, height)

        padded_rects = sf.padded_squarify(normed, x, y, width, height)

        # make plot
        if not ax or not fig:
            fig, ax = plt.subplots(figsize=(5, 4))

        ax = sf.squarify_plot_rects(padded_rects, color=clrs, labels=labels,
                                 values=label_values, ax=ax, alpha=0.8,
                                    acronyms=False)

        # TO DO: build out in next phase (model comparison)
        # if model_id:
        #     ax.set_title(f"MODEL {model_id}, {(' ').join(group_metric.split('_')).upper()} ({attribute_name.upper()})",
        #              fontsize=23, fontweight="bold")

        if title:
            ax.set_title(f"{(' ').join(related_disparity.split('_')).upper()} ({attribute_name.upper()})",
                     fontsize=23, fontweight="bold")

        if not highlight_fairness:
            # create dummy invisible image with a color map to leverage for color bar
            img = plt.imshow([[0, 2]], cmap=aq_palette, alpha=0.8)
            img.set_visible(False)
            fig.colorbar(img, orientation="vertical", shrink=.96, ax=ax)

        #     Remove axes and display the plot
        ax.axis('off')


    def plot_fairness_group(self, fairness_table, group_metric, ax=None,
                            ax_lim=None, title=False, label_dict=None,
                            min_group_size=None):
        '''
        This function plots absolute group metrics as indicated by the config file,
        colored based on calculated parity.

        :param fairness_table: a fairness table. Output of fairness.get_fairness
            function.
        :param group_metric: the fairness metric to plot. Must be a column in the fairness_table.
        :param ax: a matplotlib Axis. If not passed a new figure will be created.
        :param ax_lim: maximum value on x-axis, used to match axes across subplots
            when plotting multiple metrics. Default is None.
        :param title: whether to include a title in visualizations. Default is True.
        :param label_dict: optional dictionary of replacement values for data.
            Default is None.
        :param min_group_size: minimum proportion of total group size (all data)
            a population group must meet in order to be included in fairness
            visualization

        :return: A Matplotlib axis
        '''
        if group_metric not in fairness_table.columns:
            raise ValueError(f"Specified disparity metric {group_metric} not "
                             f"in 'fairness_table'.")

        if fairness_table[group_metric].isnull().any():
            raise ValueError(f"Cannot plot {group_metric}, has NaN values.")

        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 5))

        height_of_bar = 1
        attributes = fairness_table.attribute_name.unique()
        tick_indices = []
        next_bar_height = 0

        if min_group_size:
            if min_group_size > (fairness_table.group_size.max() / fairness_table.group_size.sum()):
                raise ValueError(f"'min_group_size' proportion specified: '{min_group_size}' "
                                f"is larger than all groups in sample.")

            min_size = min_group_size * fairness_table.group_size.sum()
            fairness_table = fairness_table.loc[fairness_table['group_size'] >= min_size]

        label_position_values = collections.deque(fairness_table[group_metric].values)


        # Lock absolute value metric plot x-axis to (0, 1)
        if not ax_lim:
            ax_lim = 1
        ax.set_xlim(0, ax_lim)

        for attribute in attributes:
            attribute_data = fairness_table.loc[
                fairness_table['attribute_name'] == attribute]
            values = attribute_data[group_metric].values
            grp_sizes = attribute_data['group_size'].values

            # apply red for "False" fairness determinations and green for "True"
            # determinations
            cb_green = '#1b7837'
            cb_red = '#a50026'
            parity = self._metric_parity_mapping[group_metric]
            parity_colors = [cb_green if val else
                              cb_red for val in attribute_data[parity]]

            # Set white text for red bars and black text for green bars
            label_colors = [(0, 0, 0, 1) if val == True else
                            (1, 1, 1, 1) for val in attribute_data[parity]]

            attribute_indices = \
                np.arange(next_bar_height, next_bar_height + attribute_data.shape[0],
                          step=height_of_bar)

            attribute_tick_location = \
                float((min(attribute_indices) + max(attribute_indices) +
                       height_of_bar)) / 2

            h_attribute = ax.barh(attribute_indices,
                                  width=values,
                                  color=parity_colors,
                                  align='edge', edgecolor='grey', alpha=0.8)

            if label_dict:
                labels = [label_dict.get(label, label) for label in
                          attribute_data['attribute_value'].values]
            else:
                labels = attribute_data['attribute_value'].values


            for y, label, value, text_color, g_size in zip(
                    attribute_indices, labels, values, label_colors,
                    grp_sizes):

                next_position = label_position_values.popleft()

                if ax_lim < 3:
                    CHAR_PLACEHOLDER = 0.03
                else:
                    CHAR_PLACEHOLDER = 0.25

                label_length = len(label) * CHAR_PLACEHOLDER
                max_val_length = 7 * CHAR_PLACEHOLDER
                indent_length = ax_lim * 0.025

                # bar long enough for label, enough space after bar for value
                if ((indent_length + label_length) < (next_position - indent_length)) and (
                        (next_position + indent_length + max_val_length) < (
                        ax_lim - indent_length)):

                    ax.text(next_position + indent_length, y + float(height_of_bar) / 2,
                            f"{value:.2f}", fontsize=12, verticalalignment='top')
                    ax.text(indent_length, y + float(height_of_bar) / 2,
                            label, fontsize=11, verticalalignment='top',
                            color=text_color)

                # case when bar too long for labels after bar, print all text in bar
                elif (next_position + indent_length + max_val_length) > (
                        ax_lim - indent_length):

                    ax.text(indent_length, y + float(height_of_bar) / 2,
                            f"{label}, {value:.2f}", fontsize=11,
                            verticalalignment='top', color=text_color)

                # case when bar too small for labels inside bar, print all text
                # after bar
                else:
                    ax.text(next_position + indent_length,
                            y + float(height_of_bar) / 2,
                            f"{label}, {value:.2f}", fontsize=12,
                            verticalalignment='top')

            tick_indices.append((attribute, attribute_tick_location))
            next_bar_height = max(attribute_indices) + 2 * height_of_bar

        ax.yaxis.set_ticks(list(map(lambda x: x[1], tick_indices)))
        ax.yaxis.set_ticklabels(list(map(lambda x: x[0], tick_indices)), fontsize=14)
        ax.set_axisbelow(True)
        ax.xaxis.grid(color='lightgray', which='major', linestyle='dashed')

        ax.set_xlabel('Absolute Metric Magnitude')

        if title:
            ax.set_title(f"{group_metric.upper()}", fontsize=20)

        return ax



    def plot_fairness_disparity(self, fairness_table, group_metric,
                                attribute_name, model_id=1, ax=None, fig=None,
                                title=True, min_group_size=None,
                                significance_alpha=0.05):
        """
        Plot disparity metrics colored based on calculated disparity.

        :param group_metric: the metric to plot. Must be a column in the disparity_table.
        :param attribute_name: which attribute to plot group_metric across.
        :param model_id: model ID number. Default is 1.
        :param ax: a matplotlib Axis. If not passed, a new figure will be created.
        :param fig: a matplotlib Figure. If not passed, a new figure will be created.
        :param title: whether to include a title in visualizations. Default is True.
        :param min_group_size: minimum proportion of total group size (all data)
            a population group must meet in order to be included in bias metric
            visualization
        :param significance_alpha: statistical significance level. Used to
            determine visual representation of significance (number of
            asterisks on treemap).
        :return: A Matplotlib axis
        """
        return self.plot_disparity(disparity_table=fairness_table,
                                   group_metric=group_metric,
                                   attribute_name=attribute_name,
                                   color_mapping=None, model_id=model_id,
                                   ax=ax, fig=fig, highlight_fairness=True,
                                   min_group_size=min_group_size, title=title,
                                   significance_alpha=significance_alpha)

    def _plot_multiple(self, data_table, plot_fcn, metrics=None, fillzeros=True,
                        title=True, ncols=3, label_dict=None, show_figure=True,
                        min_group_size=None):
        """
        This function plots bar charts of absolute metrics indicated by config
        file

        :param data_table: output of group.get_crosstabs, bias.get_disparity, or
            fairness.get_fairness functions
        :param plot_fcn: the single-metric plotting function to use for subplots
        :param metrics: which metric(s) to plot, or 'all.' If this value is
            null, will plot the following absolute metrics (or related disparity
            measures):
                - Predicted Prevalence (pprev),
                - Predicted Positive Rate (ppr),
                - False Discovery Rate (fdr),
                - False Omission Rate (for),
                - False Positive Rate (fpr),
                - False Negative Rate (fnr)
        :param fillzeros: Should null values be filled with zeros. Default is
            True.
        :param title: Whether to display a title on each plot. Default is True.
        :param ncols: The number of subplots to plot per row visualization
            figure.
            Default is 3.
        :param label_dict: Optional dictionary of label replacements. Default is
            None.
        :param show_figure: Whether to show figure (plt.show()). Default is
            True.
        :param min_group_size: Minimum proportion of total group size (all data)
            a population group must meet in order to be included in visualization

        :return: Returns a figure
        """
        if fillzeros:
            data_table = data_table.fillna(0)

        if plot_fcn in [self.plot_fairness_group, self.plot_group_metric]:
            if not metrics:
                metrics = \
                    [met for met in self.key_metrics if met in data_table.columns]

            elif metrics == 'all':
                all_abs_metrics = ('pprev', 'ppr', 'fdr', 'for', 'fpr', 'fnr',
                               'tpr', 'tnr', 'npv', 'precision')
                metrics = \
                    [met for met in all_abs_metrics if met in data_table.columns]

            ax_lim = 1

        # elif plot_fcn in [self.plot_fairness_disparity, self.plot_disparity]:
        else:
            if not metrics:
                metrics = \
                    [disp for disp in self.key_disparities if disp in data_table.columns]
            elif metrics == 'all':
                metrics = \
                    list(data_table.columns[data_table.columns.str.contains('_disparity')])

            ax_lim = min(10, self._nearest_quartile(max(data_table[metrics].max())) + 0.1)

        num_metrics = len(metrics)
        rows = math.ceil(num_metrics / ncols)
        if ncols == 1 or (num_metrics % ncols == 0):
            axes_to_remove = 0
        else:
            axes_to_remove = ncols - (num_metrics % ncols)

        if not (0 < rows <= num_metrics):
           raise ValueError (
               "Plot must have at least one row. Please update number of columns"
               " ('ncols') or check that at least one metric is specified in "
               "'metrics'.")
        if not (0 < ncols <= num_metrics):
           raise ValueError(
               "Plot must have at least one column, and no more columns than "
               "subplots. Please update number of columns ('ncols') or check "
               "that at least one metric is specified in 'metrics'.")

        total_plot_width = 25

        fig, axs = plt.subplots(nrows=rows, ncols=ncols,
                                figsize=(total_plot_width, 6 * rows),
                                sharey=True,
                                gridspec_kw={'wspace': 0.075, 'hspace': 0.25})

        # set a different metric to be plotted in each subplot
        ax_col = 0
        ax_row = 0

        for group_metric in metrics:
            if (ax_col >= ncols) and ((ax_col + 1) % ncols) == 1:
                ax_row += 1
                ax_col = 0

            if rows == 1:
                current_subplot = axs[ax_col]

            elif ncols == 1:
                current_subplot = axs[ax_row]
                ax_row += 1
            else:
                current_subplot = axs[ax_row, ax_col]

            plot_fcn(data_table, group_metric=group_metric, ax=current_subplot,
                     ax_lim=ax_lim, title=title, label_dict=label_dict,
                     min_group_size=min_group_size)
            ax_col += 1

        # disable axes not being used
        if axes_to_remove > 0:
            for i in np.arange(axes_to_remove):
                axs[-1, -(i + 1)].axis('off')

        if show_figure:
            plt.show()
        return fig


    def _plot_multiple_treemaps(self, data_table, plot_fcn, attributes=None,
                                 metrics=None, fillzeros=True, title=True,
                                 label_dict=None, highlight_fairness=False,
                                 show_figure=True, min_group_size=None,
                                significance_alpha=0.05):
        """
        This function plots treemaps of disparities indicated by config file

        :param data_table: output of bias.get_disparity, or fairness.get_fairness
            functions
        :param plot_fcn: Plotting function to use to plot individual disparity
            or fairness treemaps in grid
        :param attributes: which attributes to plot against. Must be specified
            if no metrics specified.
        :param metrics: which metric(s) to plot, or 'all.' MUST be specified if
            no attributes specified. If this value is null, the following
            absolute metrics/ related disparity measures will be plotted against
            specified attributes:
                - Predicted Prevalence (pprev),
                - Predicted Positive Rate (ppr),
                - False Discovery Rate (fdr),
                - False Omission Rate (for),
                - False Positive Rate (fpr),
                - False Negative Rate (fnr)
        :param fillzeros: Whether null values should be filled with zeros. Default
            is True.
        :param title: Whether to display a title on each plot. Default is True.
        :param label_dict: Optional dictionary of label replacements. Default is
            None.
        :param highlight_fairness: Whether to highlight treemaps by disparity
            magnitude, or by related fairness determination.
        :param show_figure: Whether to show figure (plt.show()). Default is True.
        :param min_group_size: Minimum proportion of total group size (all data)
            a population group must meet in order to be included in visualization
        :param significance_alpha: statistical significance level. Used to
            determine visual representation of significance (number of
            asterisks on treemap).

        :return: Returns a figure
        """

        if fillzeros:
            data_table = data_table.fillna(0)

        if all(v is None for v in [attributes, metrics]):
            raise ValueError("One of the following parameters must be specified: " \
                        "'attribute', 'metrics'.")

        if attributes:
            if not metrics:
                metrics = [abs_m for abs_m in self.key_metrics if
                           abs_m in data_table.columns]

            #         metrics = list(set(self.input_group_metrics) &
            # set(data_table.columns))
            elif metrics == 'all':
                all_abs_metrics = ['tpr_disparity', 'tnr_disparity', 'for_disparity',
                               'fdr_disparity', 'fpr_disparity', 'fnr_disparity',
                               'npv_disparity', 'precision_disparity',
                               'ppr_disparity', 'pprev_disparity']
                metrics = \
                    [abs_m for abs_m in all_abs_metrics if abs_m in data_table.columns]

            viz_title = \
                f"DISPARITY METRICS by {(', ').join(list(map(lambda x:x.upper(), attributes)))}"

        elif not attributes:
            attributes = list(data_table.attribute_name.unique())
            if metrics == 'all':
                all_disparities = ['tpr_disparity', 'tnr_disparity', 'for_disparity',
                               'fdr_disparity', 'fpr_disparity', 'fnr_disparity',
                               'npv_disparity', 'precision_disparity',
                               'ppr_disparity', 'pprev_disparity']
                metrics = [disparity for disparity in all_disparities if
                           disparity in data_table.columns]
            viz_title = f"{(', ').join(map(lambda x:x.upper(), metrics))} " \
                        f"ACROSS ATTRIBUTES"

        num_metrics = len(attributes) * len(metrics)
        if num_metrics > 1:
            ncols = 3
        else:
            ncols = 1

        rows = math.ceil(num_metrics / ncols)
        if ncols == 1 or (num_metrics % ncols == 0):
            axes_to_remove = 0
        else:
            axes_to_remove = ncols - (num_metrics % ncols)

        if not (0 < rows <= num_metrics):
           raise ValueError (
               "Plot must have at least one row. Please update number of columns"
               " ('ncols'), the list of metrics to be plotted ('metrics'), or "
               "the list of attributes to plot disparity metrics across.")
        if not (0 < ncols <= num_metrics):
           raise ValueError(
               "Plot must have at least one column, and no more columns than "
               "plots. Please update number of columns ('ncols'), the list of "
               "metrics to be plotted ('metrics'), or the list of attributes to "
               "plot disparity metrics across.")

        total_plot_width = 25

        fig, axs = plt.subplots(nrows=rows, ncols=ncols,
                                figsize=(total_plot_width, 8 * rows),
                                gridspec_kw={'wspace': 0.025, 'hspace': 0.5},
                                subplot_kw={'aspect': 'equal'})

        if highlight_fairness:
            mapping = None
        else:
            aq_palette = sns.diverging_palette(225, 35, sep=10, as_cmap=True)

            norm = matplotlib.colors.Normalize(vmin=0, vmax=2)
            mapping = matplotlib.cm.ScalarMappable(norm=norm, cmap=aq_palette)

        # set a different metric to be plotted in each subplot
        ax_col = 0
        ax_row = 0

        models = list(data_table.model_id.unique())

        for group_metric in metrics:
            for attr in attributes:
                if (ax_col >= ncols) and ((ax_col + 1) % ncols) == 1:
                    ax_row += 1
                    ax_col = 0

                if num_metrics == 1:
                    current_subplot = axs

                elif (num_metrics > 1) and (rows == 1):
                    current_subplot = axs[ax_col]

                elif (num_metrics > 1) and (ncols == 1):
                    current_subplot = axs[ax_row]
                    ax_row += 1
                else:
                    current_subplot = axs[ax_row, ax_col]

                plot_fcn(data_table, group_metric=group_metric,
                         attribute_name=attr, color_mapping=mapping,
                         ax=current_subplot, fig=fig, title=title,
                         label_dict=label_dict,
                         highlight_fairness=highlight_fairness,
                         min_group_size=min_group_size, significance_alpha=significance_alpha)

                ax_col += 1

        # disable axes not being used
        if axes_to_remove > 0:
            for i in np.arange(axes_to_remove):
                axs[-1, -(i + 1)].axis('off')

        plt.suptitle(f"{viz_title}", fontsize=25, fontweight="bold")

        # fig.tight_layout()

        if rows > 2:
            fig.subplots_adjust(top=0.95)
        else:
            fig.subplots_adjust(top=0.90)

        if show_figure:
            plt.show()
        return fig



    def plot_group_metric_all(self, data_table, metrics=None, fillzeros=True,
                              ncols=3, title=True, label_dict=None,
                              show_figure=True, min_group_size=None):
        """
        Plot multiple metrics at once from a fairness object table.

        :param data_table:  output of group.get_crosstabs, bias.get_disparity, or
            fairness.get_fairness functions.
        :param metrics: which metric(s) to plot, or 'all.'
            If this value is null, will plot:
                - Predicted Prevalence (pprev),
                - Predicted Positive Rate (ppr),
                - False Discovery Rate (fdr),
                - False Omission Rate (for),
                - False Positive Rate (fpr),
                - False Negative Rate (fnr)
        :param fillzeros: whether to fill null values with zeros. Default is
            True.
        :param ncols: number of subplots per row in figure. Default is 3.
        :param title: whether to display a title on each plot. Default is True.
        :param label_dict: optional dictionary of label replacements. Default is None.
        :param show_figure: whether to show figure (plt.show()). Default is True.
        :param min_group_size: minimum proportion of total group size (all data)
            a population group must meet in order to be included in group metric
            visualization.

        :return: A Matplotlib figure
        """
        return self._plot_multiple(
            data_table, plot_fcn=self.plot_group_metric, metrics=metrics,
            fillzeros=fillzeros, title=title, ncols=ncols, label_dict=label_dict,
            show_figure=show_figure, min_group_size=min_group_size)


    def plot_disparity_all(self, data_table, attributes=None, metrics=None,
                           fillzeros=True, title=True, label_dict=None, ncols=3,
                           show_figure=True, min_group_size=None,
                           significance_alpha=0.05):
        """
        Plot multiple metrics at once from a fairness object table.

        :param data_table:  output of group.get_crosstabs, bias.get_disparity, or
            fairness.get_fairness functions.
        :param attributes: which attribute(s) to plot metrics for. If this
            value is null, will plot metrics against all attributes.
        :param metrics: which metric(s) to plot, or 'all.'
            If this value is null, will plot:
                - Predicted Prevalence Disparity (pprev_disparity),
                - Predicted Positive Rate Disparity (ppr_disparity),
                - False Discovery Rate Disparity (fdr_disparity),
                - False Omission Rate Disparity (for_disparity),
                - False Positive Rate Disparity (fpr_disparity),
                - False Negative Rate Disparity (fnr_disparity)
        :param fillzeros: whether to fill null values with zeros. Default is True.
        :param title: whether to display a title on each plot. Default is True.
        :param label_dict: optional dictionary of label replacements. Default is
            None.
        :param show_figure: whether to show figure (plt.show()). Default is True.
        :param min_group_size: minimum proportion of total group size (all data)
            a population group must meet in order to be included in metric
            visualization.
        :param significance_alpha: statistical significance level. Used to
            determine visual representation of significance (number of
            asterisks on treemap).

        :return: A Matplotlib figure
        """
        return self._plot_multiple_treemaps(
            data_table, plot_fcn=self.plot_disparity, attributes=attributes,
            metrics=metrics, fillzeros=fillzeros, label_dict=label_dict,
            highlight_fairness=False, show_figure=show_figure, title=title,
            min_group_size=min_group_size, significance_alpha=significance_alpha)

    def plot_fairness_group_all(self, fairness_table, metrics=None, fillzeros=True,
                                ncols=3, title=True, label_dict=None,
                                show_figure=True, min_group_size=None):
        """
        Plot multiple metrics at once from a fairness object table.

        :param fairness_table: output of fairness.get_fairness functions.
        :param metrics: which metric(s) to plot, or 'all.'
            If this value is null, will plot:
                - Predicted Prevalence (pprev),
                - Predicted Positive Rate (ppr),
                - False Discovery Rate (fdr),
                - False Omission Rate (for),
                - False Positive Rate (fpr),
                - False Negative Rate (fnr)
        :param fillzeros: whether to fill null values with zeros. Default is True.
        :param ncols: number of subplots per row in figure. Default is 3.
        :param title: whether to display a title on each plot. Default is True.
        :param label_dict: optional dictionary of label replacements. Default is
            None.
        :param show_figure: whether to show figure (plt.show()). Default is True.
        :param min_group_size: minimum proportion of total group size (all data).
            a population group must meet in order to be included in fairness
            visualization

        :return: A Matplotlib figure
        """
        return self._plot_multiple(
            fairness_table, plot_fcn=self.plot_fairness_group, metrics=metrics,
            fillzeros=fillzeros, title=title, ncols=ncols, label_dict=label_dict,
            show_figure=show_figure, min_group_size=min_group_size)

    def plot_fairness_disparity_all(self, fairness_table, attributes=None,
                                    metrics=None, fillzeros=True, title=True,
                                    label_dict=None, show_figure=True,
                                    min_group_size=None, significance_alpha=0.05):
        """
        Plot multiple metrics at once from a fairness object table.

        :param fairness_table: output of fairness.get_fairness functions.
        :param attributes: which attribute(s) to plot metrics for. If this value is null, will plot metrics against all attributes.
        :param metrics: which metric(s) to plot, or 'all.'
            If this value is null, will plot:
                - Predicted Prevalence Disparity (pprev_disparity),
                - Predicted Positive Rate Disparity (ppr_disparity),
                - False Discovery Rate Disparity (fdr_disparity),
                - False Omission Rate Disparity (for_disparity),
                - False Positive Rate Disparity (fpr_disparity),
                - False Negative Rate Disparity (fnr_disparity)
        :param fillzeros: whether to fill null values with zeros. Default is True.
        :param title: whether to display a title on each plot. Default is True.
        :param label_dict: optional dictionary of label replacements. Default is
            None.
        :param show_figure: whether to show figure (plt.show()). Default is True.
        :param min_group_size: minimum proportion of total group size (all data)
            a population group must meet in order to be included in fairness
            visualization
        :param significance_alpha: statistical significance level. Used to
            determine visual representation of significance (number of
            asterisks on treemap)

        :return: A Matplotlib figure
        """
        return self._plot_multiple_treemaps(
            fairness_table, plot_fcn=self.plot_disparity, attributes=attributes,
            metrics=metrics, fillzeros=fillzeros, label_dict=label_dict,
            title=title, highlight_fairness=True, show_figure=show_figure,
            min_group_size=min_group_size, significance_alpha=significance_alpha)




'''
Attribution: Adapted from Uri Laserson's squaify source code for plotting treemaps in
matplotlib based on algorithm from Bruls, Huizing, van Wijk,
"Squarified Treemaps". (https://github.com/laserson/squarify)

Treemaps are flipped so smallest square is in bottom right corner. Added
function for plotting predefined rectangles. Made adjustments to treemap labeling.
'''

__author__ = "Pedro Saleiro <saleiro@uchicago.edu>, Loren Hinkson"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."

import matplotlib.pyplot as plt
import itertools

def normalize_sizes(sizes, dx, dy):
    '''
    Return list of normalized values corresponding to a list of supplied values
    and specified width (dx)/ height (dy) dimensions.

    :param sizes: Ordered (desc) list of values to be plotted as treemap squares
    :param dx: Total treemap width
    :param dy: Total treemap height
    :return: List of normalized values corresponding to original list
    '''
    total_size = sum(sizes)
    total_area = dx * dy
    sizes = map(float, sizes)
    sizes = map(lambda size: size * total_area / total_size, sizes)
    return list(sizes)


def layoutrow(sizes, x, y, dx, dy):
    '''
    Generate rectangle dimensions for each size in sizes when remaining width
    exceeds remaining height (dx >= dy). Rectangles will fill up height dy,
    and width will be determined by their area. Sizes should be pre-normalized
    with respect to dx * dy (i.e., they should be same units)

    :param sizes: Ordered (desc) ist of values normalized with respect to
        overall height (dy) and width values (dx)
    :param x: Rectangle start point on treemap x-axis
    :param y: Rectangle start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: List of dictionaries of rectangle dimensions
    '''
    covered_area = sum(sizes)
    width = covered_area / dy
    rects = []
    heights = 0
    for size in sizes:
        rects.append({'x': x, 'y': dy - (size / width), 'dx': width, 'dy': size / width})
        dy -= size / width
    return rects


def layoutcol(sizes, x, y, dx, dy):
    '''
    Generate rectangle dimensions for each size in sizes when remaining height
    exceeds remaining width (dx < dy). Rectangles will fill up width dx,
    and height will be determined by their area. Sizes should be pre-normalized
    with respect to dx * dy (i.e., they should be same units)

    :param sizes: List of values normalized with respect to overall height (dy)
        and width values (dx)
    :param x: Rectangle start point on treemap x-axis
    :param y: Rectangle start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: List of dictionaries of rectangle dimensions
    '''

    covered_area = sum(sizes)
    height = covered_area / dx
    rects = []
    for size in sizes:
        rects.append({'x': x, 'y': dy - height, 'dx': size / height, 'dy': height})
        x += size / height
    return rects


def layout(sizes, x, y, dx, dy):
    '''
    Call applicable helper function to generate dictionary of treemap rectangle
    dimensions based on relative width and height of remaining treemap area

    :param sizes: List of values normalized with respect to overall height (dy)
        and width values (dx)
    :param x: Rectangle start point on treemap x-axis
    :param y: Rectangle start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: List of dictionaries of rectangle dimensions
    '''
    return layoutrow(sizes, x, y, dx, dy) if dx >= dy else layoutcol(sizes, x, y, dx, dy)


def leftoverrow(sizes, x, y, dx, dy):
    '''
    Compute relative x-axis start, y-axis start, width, and height of overall
    treemap area when remaining width exceeds remaining height (dx >= dy)

    :param sizes: List of values normalized with respect to overall height (dy)
        and width values (dx)
    :param x: Remaining area start point on treemap x-axis
    :param y: Remaining area start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: Tuple of remaining area dimensions

    '''
    covered_area = sum(sizes)
    width = covered_area / dy
    leftover_x = x + width
    leftover_y = dy
    leftover_dx = dx - width
    leftover_dy = dy
    return (leftover_x, leftover_y, leftover_dx, leftover_dy)


def leftovercol(sizes, x, y, dx, dy):
    '''
    Compute relative x-axis start, y-axis start, width, and height of remaining
    treemap area when remaining height exceeds remaining width (dx < dy)

    :param sizes: List of values normalized with respect to overall height (dy)
        and width values (dx)
    :param x: Remaining area start point on treemap x-axis
    :param y: Remaining area start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: Tuple of remaining area dimensions
    '''
    covered_area = sum(sizes)
    height = covered_area / dx
    leftover_x = x
    leftover_y = dy
    leftover_dx = dx
    leftover_dy = dy - height
    return (leftover_x, leftover_y, leftover_dx, leftover_dy)


def leftover(sizes, x, y, dx, dy):
    '''
    Call applicable helper function to generate tuple of dimensions for
    remaining plottable area for rectangles

    :param sizes: List of values normalized with respect to overall height (dy)
        and width values (dx)
    :param x: Remaining area start point on treemap x-axis
    :param y: Remaining area start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: Tuple of remaining area dimensions
    '''
    return leftoverrow(sizes, x, y, dx, dy) if dx >= dy else leftovercol(sizes, x, y, dx, dy)


def worst_ratio(sizes, x, y, dx, dy):
    '''
    Calculate worst possible ratio between width (dx) and height (dy) to
    determine ideal layout for a given list of rectangle dimensions

    :param sizes: List of values normalized with respect to overall height (dy)
        and width values (dx)
    :param x: Remaining area start point on treemap x-axis
    :param y: Remaining area start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: Numeric value indicating worst ratio between height/width
        dimensions
    '''
    return max(
        max(
            rect['dx'] / rect['dy'],
            rect['dy'] / rect['dx']
            )
        for rect in layout(sizes, x, y, dx, dy)
    )


def squarify(sizes, x, y, dx, dy):
    """
    Calculate rectangle dimensions for list of values relative to a given
    width (dx) and height (dy), starting at given x-axis and y-axis values.
    List 'sizes' must be normalized, unless dx * dy == sum(sizes).

    :param sizes: Ordered (desc) list of values normalized with respect to
        overall height (dy) and width values (dx)
    :param x: Remaining area start point on treemap x-axis
    :param y: Remaining area start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height

    :return: List of dictionaries of rectangle dimensions
    """
    sizes = list(map(float, sizes))

    if len(sizes) == 0:
        return []

    if len(sizes) == 1:
        return layout(sizes, x, y, dx, dy)

    # figure out where 'split' should be based on utilization of remaining area
    # determined by ratio between width (dx) and height (dy)
    i = 1
    while i < len(sizes) and worst_ratio(sizes[:i], x, y, dx, dy) >= worst_ratio(sizes[:(i + 1)], x, y, dx, dy):
        i += 1
    current = sizes[:i]

    remaining = sizes[i:]

    (leftover_x, leftover_y, leftover_dx, leftover_dy) = leftover(current, x, y, dx, dy)

    return layout(current, x, y, dx, dy) + \
       squarify(remaining, leftover_x, leftover_y, leftover_dx, leftover_dy)


def pad_rectangle(rect):
    '''
    Decrease rectangle dimensions to show whitespace in treemap

    :param rect: Dictionary of rectangle dimensions to be decreased to enable
        whitespace padding in treemap visualization
    :return: Dictionary of updated rectangle dimensions
    '''
    if rect['dx'] > 2:
        rect['x'] += 1
        rect['dx'] -= 2
    if rect['dy'] > 2:
        rect['y'] += 1
        rect['dy'] -= 2


def padded_squarify(sizes, x, y, dx, dy):
    '''
    Calculate rectangle dimensions relative to a given width (dx) and height (dy)
    starting at given x-axis and y-axis values, leaving room for whitespace
    padding between treemap rectangles


    :param sizes: Ordered (desc) list of values normalized with respect to
        overall height (dy) and width values (dx)
    :param x: Remaining area start point on treemap x-axis
    :param y: Remaining area start point on treemap y-axis
    :param dx: Remaining treemap width
    :param dy: Remaining treemap height
    :return: List of dictionaries of rectangle dimensions

    '''
    rects = squarify(sizes, x, y, dx, dy)
    for rect in rects:
        pad_rectangle(rect)
    return rects

def squarify_plot_rects(rects, norm_x=100, norm_y=100, color=None,
                        labels=None, values=None, ax=None, acronyms = False, **kwargs):
    """
    Plotting with Matplotlib from predefined rectangles. Adapted from squarify
    source code.

    :param rects: list-like of dictionaries indicating rectangle dimensions
        for plotting
    :param norm_x:  overall figure dimensions for normalizing value box sizes
    :param norm_y: overall figure dimensions for normalizing value box sizes
    :param color: color string or list-like of colors to use for value boxes
    :param labels: list-like used as label text
    :param values: list-like used as value text
    :param ax: Matplotlib Axes instance
    :param kwargs: dict, keyword arguments passed to matplotlib.Axes.bar

    :return: matplotlib.Axis
    """
    x = [rect['x'] for rect in rects]
    y = [rect['y'] for rect in rects]
    dx = [rect['dx'] for rect in rects]
    dy = [rect['dy'] for rect in rects]

    ax.bar(x, dy, width=dx, bottom=y, color=color,
           label=labels, align='edge', **kwargs)

    INDENT_LENGTH = 4
    CHAR_PLACEHOLDER = 1.5

    if values is not None:
        va = 'center' if labels is None else 'top'
        if values:
            plot_ready_values = [
                val if isinstance(val, str)
                else f"{val:.2}" if isinstance(val, (int, float))
                else ""
                for val in values
            ]

        for val, r in zip(plot_ready_values, rects):
            x, y, dx, dy = r['x'], r['y'], r['dx'], r['dy']

            # if box large enough, add labels and values
            if (dx >= (INDENT_LENGTH * 2) + (CHAR_PLACEHOLDER * len(val))) &\
                    (dx > 10):
                ax.text(x + dx / 2, y + dy / 2, val, va=va,
                        ha='center', fontsize=12)

    if labels is not None:
        va = 'center' if values is None else 'bottom'
        under_plot = []
        alphabet = list(map(chr, range(65, 91)))
        under_plot_num = 0

        if values:
            plot_ready_values = [
                val if isinstance(val, str)
                else f"{val:.2}" if isinstance(val, (int, float))
                else ""
                for val in values
            ]
        else:
            plot_ready_values = itertools.repeat(None)

        for (label, r, val) in zip(labels, rects, plot_ready_values):
            x, y, dx, dy = r['x'], r['y'], r['dx'], r['dy']
            length = dx

            # if box large enough, add labels and values
            if (dx >= (INDENT_LENGTH * 2) + CHAR_PLACEHOLDER * len(label)) & (dx > 10):
                ax.text(x + dx / 2, y + dy / 2, label, va=va, ha='center',
                        fontsize=14, wrap=False)


            else:
                # add labels that don't fit in boxes underneath plot
                if acronyms:
                    # use acronym to to label very small boxes
                    acronym = ''.join([word[0] for word in str(label).split(' ')])

                    # add labels that don't fit in boxes underneath plot
                    ax.text(x + dx / 2, y + dy / 2, acronym, va=va,
                            ha='center', fontsize=12, wrap=False)

                    underplot_label = str(label) if val is None else f"{label}, {val}"
                    under_plot.append(f"{acronym}: {underplot_label}")

                else:
                    # use alphabet character to to label very small boxes
                    ax.text(x + dx / 2, y + dy / 2, alphabet[under_plot_num], va=va,
                            ha='center', fontsize=12, wrap=False)

                    underplot_label = str(label) if val is None else f"{label}, {val}"
                    under_plot.append(f"{alphabet[under_plot_num]}: {underplot_label}")
                    under_plot_num += 1


        if len(under_plot) > 0:
            unlabeled = ('\n').join(under_plot)
            ax.text(0.0, -0.05, f"Not labeled above:\n{unlabeled}",
                    ha='left', va='top', transform=ax.transAxes, fontsize=14)

    ax.set_xlim(0, norm_x)
    ax.set_ylim(0, norm_y)
    return ax
